# pylint: disable=duplicate-code
# -*- coding: utf-8 -*-
#
#       ramstk.models.design_electric.record.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTKDesignElectric Table Module."""

# Standard Library Imports
from typing import Any, Dict, List, Tuple

# Third Party Imports
from pubsub import pub
from sqlalchemy import Column, Float, ForeignKey, Integer, String

# RAMSTK Package Imports
from ramstk.analyses import derating, stress
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord


def do_check_overstress(
    overstress: Dict[str, List[float]], stress_type: str
) -> Tuple[bool, str]:
    """Check the overstress condition and build a reason message.

    :param overstress: the dict containing the results of the
        overstress analysis.
    :param stress_type: the overstress type being checked.
    :return: (_overstress, _reason); whether a component is overstressed and the reason.
    :rtype: tuple
    """
    _overstress = False
    _reason = ""

    if overstress["harsh"][0]:
        _overstress = True
        _reason = _reason + (
            f"Operating {stress_type} is less than limit in a harsh environment.\n"
        )
    if overstress["harsh"][1]:
        _overstress = True
        _reason = _reason + (
            f"Operating {stress_type} is greater than limit "
            f"in a harsh environment.\n"
        )
    if overstress["mild"][0]:
        _overstress = True
        _reason = _reason + (
            f"Operating {stress_type} is less than limit in a mild environment.\n"
        )
    if overstress["mild"][1]:
        _overstress = True
        _reason = _reason + (
            f"Operating {stress_type} is greater than limit "
            f"in a mild environment.\n"
        )

    return _overstress, _reason


# pylint: disable=R0902
class RAMSTKDesignElectricRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    """Represent ramstk_design_electric table in the RAMSTK Program database.

    This table shares a One-to-One relationship with ramstk_hardware.
    """

    __defaults__ = {
        "application_id": 0,
        "area": 0.0,
        "capacitance": 0.0,
        "configuration_id": 0,
        "construction_id": 0,
        "contact_form_id": 0,
        "contact_gauge": 0,
        "contact_rating_id": 0,
        "current_operating": 0.0,
        "current_rated": 0.0,
        "current_ratio": 0.0,
        "environment_active_id": 0,
        "environment_dormant_id": 0,
        "family_id": 0,
        "feature_size": 0.0,
        "frequency_operating": 0.0,
        "insert_id": 0,
        "insulation_id": 0,
        "manufacturing_id": 0,
        "matching_id": 0,
        "n_active_pins": 0,
        "n_circuit_planes": 1,
        "n_cycles": 0,
        "n_elements": 0,
        "n_hand_soldered": 0,
        "n_wave_soldered": 0,
        "operating_life": 0.0,
        "overstress": 0,
        "package_id": 0,
        "power_operating": 0.0,
        "power_rated": 0.0,
        "power_ratio": 0.0,
        "reason": "",
        "resistance": 0.0,
        "specification_id": 0,
        "technology_id": 0,
        "temperature_active": 35.0,
        "temperature_case": 0.0,
        "temperature_dormant": 25.0,
        "temperature_hot_spot": 0.0,
        "temperature_junction": 0.0,
        "temperature_knee": 25.0,
        "temperature_rated_max": 0.0,
        "temperature_rated_min": 0.0,
        "temperature_rise": 0.0,
        "theta_jc": 0.0,
        "type_id": 0,
        "voltage_ac_operating": 0.0,
        "voltage_dc_operating": 0.0,
        "voltage_esd": 0.0,
        "voltage_rated": 0.0,
        "voltage_ratio": 0.0,
        "weight": 0.0,
        "years_in_production": 1,
    }
    __tablename__ = "ramstk_design_electric"
    __table_args__ = {"extend_existing": True}

    revision_id = Column(
        "fld_revision_id",
        Integer,
        ForeignKey("ramstk_revision.fld_revision_id", ondelete="CASCADE"),
        nullable=False,
    )
    hardware_id = Column(
        "fld_hardware_id",
        Integer,
        ForeignKey("ramstk_hardware.fld_hardware_id", ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )

    application_id = Column(
        "fld_application_id", Integer, default=__defaults__["application_id"]
    )
    area = Column("fld_area", Float, default=__defaults__["area"])
    capacitance = Column("fld_capacitance", Float, default=__defaults__["capacitance"])
    configuration_id = Column(
        "fld_configuration_id", Integer, default=__defaults__["configuration_id"]
    )
    construction_id = Column(
        "fld_construction_id", Integer, default=__defaults__["construction_id"]
    )
    contact_form_id = Column(
        "fld_contact_form_id", Integer, default=__defaults__["contact_form_id"]
    )
    contact_gauge = Column(
        "fld_contact_gauge", Integer, default=__defaults__["contact_gauge"]
    )
    contact_rating_id = Column(
        "fld_contact_rating_id", Integer, default=__defaults__["contact_rating_id"]
    )
    current_operating = Column(
        "fld_current_operating", Float, default=__defaults__["current_operating"]
    )
    current_rated = Column(
        "fld_current_rated", Float, default=__defaults__["current_rated"]
    )
    current_ratio = Column(
        "fld_current_ratio", Float, default=__defaults__["current_ratio"]
    )
    environment_active_id = Column(
        "fld_environment_active_id",
        Integer,
        default=__defaults__["environment_active_id"],
    )
    environment_dormant_id = Column(
        "fld_environment_dormant_id",
        Integer,
        default=__defaults__["environment_dormant_id"],
    )
    family_id = Column("fld_family_id", Integer, default=__defaults__["family_id"])
    feature_size = Column(
        "fld_feature_size", Float, default=__defaults__["feature_size"]
    )
    frequency_operating = Column(
        "fld_frequency_operating", Float, default=__defaults__["frequency_operating"]
    )
    insert_id = Column("fld_insert_id", Integer, default=__defaults__["insert_id"])
    insulation_id = Column(
        "fld_insulation_id", Integer, default=__defaults__["insulation_id"]
    )
    manufacturing_id = Column(
        "fld_manufacturing_id", Integer, default=__defaults__["manufacturing_id"]
    )
    matching_id = Column(
        "fld_matching_id", Integer, default=__defaults__["matching_id"]
    )
    n_active_pins = Column(
        "fld_n_active_pins", Integer, default=__defaults__["n_active_pins"]
    )
    n_circuit_planes = Column(
        "fld_n_circuit_planes", Integer, default=__defaults__["n_circuit_planes"]
    )
    n_cycles = Column("fld_n_cycles", Integer, default=__defaults__["n_cycles"])
    n_elements = Column("fld_n_elements", Integer, default=__defaults__["n_elements"])
    n_hand_soldered = Column(
        "fld_n_hand_soldered", Integer, default=__defaults__["n_hand_soldered"]
    )
    n_wave_soldered = Column(
        "fld_n_wave_soldered", Integer, default=__defaults__["n_wave_soldered"]
    )
    operating_life = Column(
        "fld_operating_life", Float, default=__defaults__["operating_life"]
    )
    overstress = Column("fld_overstress", Integer, default=__defaults__["overstress"])
    package_id = Column("fld_package_id", Integer, default=__defaults__["package_id"])
    power_operating = Column(
        "fld_power_operating", Float, default=__defaults__["power_operating"]
    )
    power_rated = Column("fld_power_rated", Float, default=__defaults__["power_rated"])
    power_ratio = Column("fld_power_ratio", Float, default=__defaults__["power_ratio"])
    reason = Column("fld_reason", String, default=__defaults__["reason"])
    resistance = Column("fld_resistance", Float, default=__defaults__["resistance"])
    specification_id = Column(
        "fld_specification_id", Integer, default=__defaults__["specification_id"]
    )
    technology_id = Column(
        "fld_technology_id", Integer, default=__defaults__["technology_id"]
    )
    temperature_active = Column(
        "fld_temperature_active", Float, default=__defaults__["temperature_active"]
    )
    temperature_case = Column(
        "fld_temperature_case", Float, default=__defaults__["temperature_case"]
    )
    temperature_dormant = Column(
        "fld_temperature_dormant", Float, default=__defaults__["temperature_dormant"]
    )
    temperature_hot_spot = Column(
        "fld_temperature_hot_spot", Float, default=__defaults__["temperature_hot_spot"]
    )
    temperature_junction = Column(
        "fld_temperature_junction", Float, default=__defaults__["temperature_junction"]
    )
    temperature_knee = Column(
        "fld_temperature_knee", Float, default=__defaults__["temperature_knee"]
    )
    temperature_rated_max = Column(
        "fld_temperature_rated_max",
        Float,
        default=__defaults__["temperature_rated_max"],
    )
    temperature_rated_min = Column(
        "fld_temperature_rated_min",
        Float,
        default=__defaults__["temperature_rated_min"],
    )
    temperature_rise = Column(
        "fld_temperature_rise", Float, default=__defaults__["temperature_rise"]
    )
    theta_jc = Column("fld_theta_jc", Float, default=__defaults__["theta_jc"])
    type_id = Column("fld_type_id", Integer, default=__defaults__["type_id"])
    voltage_ac_operating = Column(
        "fld_voltage_ac_operating", Float, default=__defaults__["voltage_ac_operating"]
    )
    voltage_dc_operating = Column(
        "fld_voltage_dc_operating", Float, default=__defaults__["voltage_dc_operating"]
    )
    voltage_esd = Column("fld_voltage_esd", Float, default=__defaults__["voltage_esd"])
    voltage_rated = Column(
        "fld_voltage_rated", Float, default=__defaults__["voltage_rated"]
    )
    voltage_ratio = Column(
        "fld_voltage_ratio", Float, default=__defaults__["voltage_ratio"]
    )
    weight = Column("fld_weight", Float, default=__defaults__["weight"])
    years_in_production = Column(
        "fld_years_in_production", Integer, default=__defaults__["years_in_production"]
    )

    # Define the relationships to other tables in the RAMSTK Program database.

    def get_attributes(self) -> Dict[str, Any]:
        """Retrieve current values of RAMSTKDesignElectric model attributes.

        :return: {hardware_id, application_id, area, capacitance,
                  configuration_id, construction_id, contact_form_id,
                  contact_gauge, contact_rating_id, current_operating,
                  current_rated, current_ratio, environment_active_id,
                  environment_dormant_id, family_id, feature_size,
                  frequency_operating, insert_id, insulation_id,
                  manufacturing_id, matching_id, n_active_pins,
                  n_circuit_planes, n_cycles, n_elements, n_hand_soldered,
                  n_wave_soldered, operating_life, overstress, package_id,
                  power_operating, power_rated, power_ratio, reason,
                  resistance, specification_id, technology_id,
                  temperature_active, temperature_case, temperature_dormant,
                  temperature_hot_spot, temperature_junction, temperature_knee,
                  temperature_rated_max, temperature_rated_min,
                  temperature_rise, theta_jc, type_id, voltage_ac_operating,
                  voltage_dc_operating, voltage_esd, voltage_rated,
                  voltage_ratio, weight, years_in_production} pairs.
        :rtype: dict
        """
        _attributes = {
            "hardware_id": self.hardware_id,
            "application_id": self.application_id,
            "area": self.area,
            "capacitance": self.capacitance,
            "configuration_id": self.configuration_id,
            "construction_id": self.construction_id,
            "contact_form_id": self.contact_form_id,
            "contact_gauge": self.contact_gauge,
            "contact_rating_id": self.contact_rating_id,
            "current_operating": self.current_operating,
            "current_rated": self.current_rated,
            "current_ratio": self.current_ratio,
            "environment_active_id": self.environment_active_id,
            "environment_dormant_id": self.environment_dormant_id,
            "family_id": self.family_id,
            "feature_size": self.feature_size,
            "frequency_operating": self.frequency_operating,
            "insert_id": self.insert_id,
            "insulation_id": self.insulation_id,
            "manufacturing_id": self.manufacturing_id,
            "matching_id": self.matching_id,
            "n_active_pins": self.n_active_pins,
            "n_circuit_planes": self.n_circuit_planes,
            "n_cycles": self.n_cycles,
            "n_elements": self.n_elements,
            "n_hand_soldered": self.n_hand_soldered,
            "n_wave_soldered": self.n_wave_soldered,
            "operating_life": self.operating_life,
            "overstress": self.overstress,
            "package_id": self.package_id,
            "power_operating": self.power_operating,
            "power_rated": self.power_rated,
            "power_ratio": self.power_ratio,
            "reason": self.reason,
            "resistance": self.resistance,
            "specification_id": self.specification_id,
            "technology_id": self.technology_id,
            "temperature_active": self.temperature_active,
            "temperature_case": self.temperature_case,
            "temperature_dormant": self.temperature_dormant,
            "temperature_hot_spot": self.temperature_hot_spot,
            "temperature_junction": self.temperature_junction,
            "temperature_knee": self.temperature_knee,
            "temperature_rated_max": self.temperature_rated_max,
            "temperature_rated_min": self.temperature_rated_min,
            "temperature_rise": self.temperature_rise,
            "theta_jc": self.theta_jc,
            "type_id": self.type_id,
            "voltage_ac_operating": self.voltage_ac_operating,
            "voltage_dc_operating": self.voltage_dc_operating,
            "voltage_esd": self.voltage_esd,
            "voltage_rated": self.voltage_rated,
            "voltage_ratio": self.voltage_ratio,
            "weight": self.weight,
            "years_in_production": self.years_in_production,
        }

        return _attributes

    def do_calculate_current_ratio(self) -> None:
        """Calculate the current ratio.

        :return: None
        :rtype: None
        """
        try:
            self.current_ratio = stress.calculate_stress_ratio(
                self.current_operating,
                self.current_rated,
            )
        except ZeroDivisionError:
            _error_msg: str = (
                f"Failed to calculate current ratio for hardware ID "
                f"{self.hardware_id}.  Rated current={self.current_rated}, operating "
                f"current={self.current_operating}."
            )
            pub.sendMessage(
                "do_log_debug",
                logger_name="DEBUG",
                message=_error_msg,
            )
            pub.sendMessage(
                "fail_calculate_current_stress",
                error_message=_error_msg,
            )

    def do_calculate_power_ratio(self) -> None:
        """Calculate the power ratio.

        :return: None
        :rtype: None
        """
        try:
            self.power_ratio = stress.calculate_stress_ratio(
                self.power_operating,
                self.power_rated,
            )
        except ZeroDivisionError:
            _error_msg: str = (
                f"Failed to calculate power ratio for hardware ID {self.hardware_id}.  "
                f"Rated power={self.power_rated}, operating "
                f"power={self.power_operating}."
            )
            pub.sendMessage(
                "do_log_debug",
                logger_name="DEBUG",
                message=_error_msg,
            )
            pub.sendMessage(
                "fail_calculate_power_stress",
                error_message=_error_msg,
            )

    def do_calculate_voltage_ratio(self) -> None:
        """Calculate the voltage ratio.

        :return: None
        :rtype: None
        """
        _voltage_operating = self.voltage_ac_operating + self.voltage_dc_operating

        try:
            self.voltage_ratio = stress.calculate_stress_ratio(
                _voltage_operating, self.voltage_rated
            )
        except ZeroDivisionError:
            _error_msg: str = (
                f"Failed to calculate voltage ratio for hardware ID "
                f"{self.hardware_id}.  Rated voltage={self.voltage_rated}, "
                f"operating ac voltage={self.voltage_ac_operating}, operating DC "
                f"voltage={self.voltage_dc_operating}."
            )
            pub.sendMessage(
                "do_log_debug",
                logger_name="DEBUG",
                message=_error_msg,
            )
            pub.sendMessage(
                "fail_calculate_voltage_stress",
                error_message=_error_msg,
            )

    def do_derating_analysis(self, stress_limits: List[float]) -> None:
        """Perform a derating analysis.

        :param stress_limits: the list of stress limits for the selected record.
        :return: None
        :rtype: None
        """
        _overstress = False
        _reason = ""

        _current_limits = {
            "harsh": [0.0, stress_limits[0]],
            "mild": [0.0, stress_limits[1]],
        }
        _power_limits = {
            "harsh": [0.0, stress_limits[2]],
            "mild": [0.0, stress_limits[3]],
        }
        _voltage_limits = {
            "harsh": [0.0, stress_limits[4]],
            "mild": [0.0, stress_limits[5]],
        }

        _ostress, _rsn = do_check_overstress(
            derating.check_overstress(self.current_ratio, _current_limits),
            "current",
        )
        _overstress = _overstress or _ostress
        _reason = _reason + _rsn

        _ostress, _rsn = do_check_overstress(
            derating.check_overstress(self.power_ratio, _power_limits), "power"
        )
        _overstress = _overstress or _ostress
        _reason = _reason + _rsn

        _ostress, _rsn = do_check_overstress(
            derating.check_overstress(self.voltage_ratio, _voltage_limits), "voltage"
        )
        self.overstress = _overstress or _ostress
        self.reason = _reason + _rsn

    def do_stress_analysis(self, category_id: int) -> None:
        """Perform a stress analysis.

        :param category_id: the component category ID of the record to calculate.
        :return: None
        :rtype: None
        """
        if category_id in [1, 2, 5, 6, 7, 8]:
            self.do_calculate_current_ratio()

        if category_id == 3:
            self.do_calculate_power_ratio()

        if category_id in [4, 5, 8]:
            self.do_calculate_voltage_ratio()



# -*- coding: utf-8 -*-
#
#       ramstk.models.commondb.RAMSTKType.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2017 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTKType Table Module."""

# Third Party Imports
from sqlalchemy import Column, Integer, String

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord


class RAMSTKType(RAMSTK_BASE, RAMSTKBaseRecord):
    """Class to represent tramstk_type in the RAMSTK Common database."""

    __defaults__ = {
        "code": "Type Code",
        "description": "Type Description",
        "type_type": "unknown",
    }
    __tablename__ = "ramstk_type"
    __table_args__ = {"extend_existing": True}

    type_id = Column(
        "fld_type_id",
        Integer,
        primary_key=True,
        autoincrement=True,
        nullable=False,
    )
    code = Column("fld_code", String(256), default=__defaults__["code"])
    description = Column(
        "fld_description", String(512), default=__defaults__["description"]
    )
    type_type = Column("fld_type", String(256), default=__defaults__["type_type"])

    def get_attributes(self):
        """Retrieve the current values of the RAMSTKType data model attributes.

        :return: {type_id, description, type_type} pairs.
        :rtype: dict
        """
        _attributes = {
            "type_id": self.type_id,
            "code": self.code,
            "description": self.description,
            "type_type": self.type_type,
        }

        return _attributes



# Standard Library Imports
from typing import Any, Dict, List, TextIO, Tuple

# Third Party Imports
from sqlalchemy.engine import Engine
from sqlalchemy.orm import query as query
from sqlalchemy.orm import scoped_session

# RAMSTK Package Imports
from ramstk.exceptions import DataAccessError as DataAccessError

def do_create_program_db(database: Dict[str, str], sql_file: TextIO) -> None: ...
def do_open_session(database: str) -> Tuple[Engine, scoped_session]: ...

class BaseDatabase:
    cxnargs: Dict[str, str] = ...
    engine: Engine = ...
    session: scoped_session = ...
    database: str = ...
    sqlstatements: Dict[str, str] = ...
    def __init__(self) -> None: ...
    def do_connect(self, database: Dict) -> None: ...
    def do_delete(self, item: object) -> None: ...
    def do_disconnect(self) -> None: ...
    def do_insert(self, record: object) -> None: ...
    def do_insert_many(self, records: List[object]) -> None: ...
    def do_select_all(self, table: Any, **kwargs: Any) -> query.Query: ...
    def do_update(self, record: object = ...) -> None: ...
    def get_database_list(self, database: Dict[str, str]) -> List: ...
    def get_last_id(self, table: str, id_column: str) -> Any: ...



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.assistants.project.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2020 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""The RAMSTK Open Project Assistant Module."""

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration
from ramstk.db.base import BaseDatabase
from ramstk.views.gtk3 import Gtk, _
from ramstk.views.gtk3.widgets.dialog import (
    RAMSTKDatabaseSelect,
    RAMSTKDialog,
    RAMSTKMessageDialog,
)
from ramstk.views.gtk3.widgets.label import RAMSTKLabel


class CreateProject:
    """The class used to create a new RAMSTK Project database."""

    RAMSTK_USER_CONFIGURATION: RAMSTKUserConfiguration = None  # type: ignore

    def __init__(
        self,
        __button: Gtk.ToolButton,
        configuration: RAMSTKUserConfiguration,
        parent: object,
    ) -> None:
        """Initialize an instance of the Create Project Assistant.

        :param __button: the Gtk.ToolButton() that launched this class.
        :param configuration: the RAMSTKUserConfiguration class instance.
        :param parent: the parent window associated with the dialog.
        """
        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.
        self._parent: object = parent

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.RAMSTK_USER_CONFIGURATION = configuration

        self._do_request_create_project()

    def _do_confirm_overwrite(self, database: str) -> None:
        """Raise dialog to confirm overwriting existing RAMSTK database.

        :param database: the name of the existing database that is to be
            confirmed for overwrite.
        :return: None
        :rtype: None
        """
        _dialog = RAMSTKDialog(
            _("RAMSTK - Confirm Overwrite"),
            dlgbuttons=(
                Gtk.STOCK_YES,
                Gtk.ResponseType.YES,
                Gtk.STOCK_NO,
                Gtk.ResponseType.NO,
            ),
            dlgparent=self._parent,
        )

        _label = RAMSTKLabel(
            _(
                "RAMSTK Program database already exists:" "\n\n\t\t{0:s}\n\nOverwrite?"
            ).format(database)
        )
        _label.do_set_properties(width=-1, height=-1, bold=False, wrap=True)
        _dialog.vbox.pack_start(_label, True, True, 0)
        _dialog.show_all()

        if _dialog.run() == Gtk.ResponseType.YES:
            self.RAMSTK_USER_CONFIGURATION.RAMSTK_PROG_INFO["database"] = database
        else:
            self.RAMSTK_USER_CONFIGURATION.RAMSTK_PROG_INFO["database"] = ""

        _dialog.destroy()

    def _do_request_create_project(self) -> None:
        """Request to create a new RAMSTK Project Database.

        :return: None
        :rtype: None
        """
        _dialog = RAMSTKDatabaseSelect(
            dlgtitle=(
                "Select RAMSTK Program "
                "Database on the {0:s} Server".format(
                    self.RAMSTK_USER_CONFIGURATION.RAMSTK_PROG_INFO["dialect"]
                )
            ),
            dlgparent=self._parent,
            dao=BaseDatabase(),
            database=self.RAMSTK_USER_CONFIGURATION.RAMSTK_PROG_INFO,
            icons={
                "refresh": self.RAMSTK_USER_CONFIGURATION.RAMSTK_CONF_DIR
                + "/icons/32x32/view-refresh.png",
                "save": self.RAMSTK_USER_CONFIGURATION.RAMSTK_CONF_DIR
                + "/icons/32x32/save.png",
            },
        )

        if _dialog.do_run() == Gtk.ResponseType.OK:
            self.RAMSTK_USER_CONFIGURATION.RAMSTK_PROG_INFO = _dialog.database

            if _dialog.exists:
                self._do_confirm_overwrite(_dialog.database["database"])

            pub.sendMessage(
                "request_create_program",
                program_db=BaseDatabase(),
                database=self.RAMSTK_USER_CONFIGURATION.RAMSTK_PROG_INFO,
            )

        _dialog.destroy()


class OpenProject:
    """Assistant to guide user through process of creating RAMSTK Project."""

    RAMSTK_USER_CONFIGURATION: RAMSTKUserConfiguration = None  # type: ignore

    def __init__(
        self,
        __button: Gtk.ToolButton,
        configuration: RAMSTKUserConfiguration,
        parent: object,
    ) -> None:
        """Initialize an instance of the Create Project Assistant.

        :param __button: the Gtk.ToolButton() that launched an instance of this
            class.
        :param configuration: the RAMSTKUserConfiguration class instance.
        """
        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.
        self._parent: object = parent

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.RAMSTK_USER_CONFIGURATION = configuration

        self._do_request_open_project()

    def _do_request_open_project(self) -> None:
        """Open or connect to a RAMSTK Program database.

        :return: None
        :rtype: None
        """
        if self.RAMSTK_USER_CONFIGURATION.loaded:
            self.__project_is_open()
        else:
            _dialog = RAMSTKDatabaseSelect(
                dlgtitle=(
                    "Select RAMSTK Program "
                    "Database on the {0:s} Server".format(
                        self.RAMSTK_USER_CONFIGURATION.RAMSTK_PROG_INFO["dialect"]
                    )
                ),
                dlgparent=self._parent,
                dao=BaseDatabase(),
                database=self.RAMSTK_USER_CONFIGURATION.RAMSTK_PROG_INFO,
                icons={
                    "refresh": self.RAMSTK_USER_CONFIGURATION.RAMSTK_CONF_DIR
                    + "/icons/32x32/view-refresh.png",
                    "save": self.RAMSTK_USER_CONFIGURATION.RAMSTK_CONF_DIR
                    + "/icons/32x32/save.png",
                },
            )

            if _dialog.do_run() == Gtk.ResponseType.OK:
                self.RAMSTK_USER_CONFIGURATION.RAMSTK_PROG_INFO = _dialog.database

                if _dialog.exists:
                    pub.sendMessage(
                        "request_open_program",
                        program_db=BaseDatabase(),
                        database=self.RAMSTK_USER_CONFIGURATION.RAMSTK_PROG_INFO,
                    )
                else:
                    pub.sendMessage(
                        "request_create_program",
                        program_db=BaseDatabase(),
                        database=self.RAMSTK_USER_CONFIGURATION.RAMSTK_PROG_INFO,
                    )

                if _dialog.btnSave.get_active():
                    self.RAMSTK_USER_CONFIGURATION.set_user_configuration()

            _dialog.do_destroy()

    def __project_is_open(self) -> None:
        """Raise dialog explaining a project is already open.

        :return: None
        """
        _prompt = _(
            "A database is already open.  Only one database can "
            "be open at a time in RAMSTK.  You must close the "
            "currently open RAMSTK database before a new "
            "database can be opened."
        )
        _dialog = RAMSTKMessageDialog(parent=self._parent)
        _dialog.do_set_message(_prompt)
        _dialog.do_set_message_type("information")

        if _dialog.run() == Gtk.ResponseType.OK:
            _dialog.destroy()



# pylint: disable=unused-import
# -*- coding: utf-8 -*-
#
#       ramstk.models.__init__.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""The RAMSTK database models package."""

from .basemodel import RAMSTKBaseRecord, RAMSTKBaseTable, RAMSTKBaseView  # isort:skip
from .commondb.site_info.record import RAMSTKSiteInfoRecord  # isort:skip

# RAMSTK Local Imports
from .commondb.database import RAMSTKCommonDB
from .commondb.site_info.table import RAMSTKSiteInfoTable
from .programdb.action.record import RAMSTKActionRecord
from .programdb.action.table import RAMSTKActionTable
from .programdb.allocation.record import RAMSTKAllocationRecord
from .programdb.allocation.table import RAMSTKAllocationTable
from .programdb.cause.record import RAMSTKCauseRecord
from .programdb.cause.table import RAMSTKCauseTable
from .programdb.control.record import RAMSTKControlRecord
from .programdb.control.table import RAMSTKControlTable
from .programdb.database import RAMSTKProgramDB
from .programdb.design_electric.record import RAMSTKDesignElectricRecord
from .programdb.design_electric.table import RAMSTKDesignElectricTable
from .programdb.design_mechanic.record import RAMSTKDesignMechanicRecord
from .programdb.design_mechanic.table import RAMSTKDesignMechanicTable
from .programdb.environment.record import RAMSTKEnvironmentRecord
from .programdb.environment.table import RAMSTKEnvironmentTable
from .programdb.failure_definition.record import RAMSTKFailureDefinitionRecord
from .programdb.failure_definition.table import RAMSTKFailureDefinitionTable
from .programdb.fmea.view import RAMSTKFMEAView
from .programdb.function.record import RAMSTKFunctionRecord
from .programdb.function.table import RAMSTKFunctionTable
from .programdb.hardware.record import RAMSTKHardwareRecord
from .programdb.hardware.table import RAMSTKHardwareTable
from .programdb.hardware.view import RAMSTKHardwareBoMView
from .programdb.hazard.record import RAMSTKHazardRecord
from .programdb.hazard.table import RAMSTKHazardTable
from .programdb.mechanism.record import RAMSTKMechanismRecord
from .programdb.mechanism.table import RAMSTKMechanismTable
from .programdb.milhdbk217f.record import RAMSTKMilHdbk217FRecord
from .programdb.milhdbk217f.table import RAMSTKMILHDBK217FTable
from .programdb.mission.record import RAMSTKMissionRecord
from .programdb.mission.table import RAMSTKMissionTable
from .programdb.mission_phase.record import RAMSTKMissionPhaseRecord
from .programdb.mission_phase.table import RAMSTKMissionPhaseTable
from .programdb.mode.record import RAMSTKModeRecord
from .programdb.mode.table import RAMSTKModeTable
from .programdb.nswc.record import RAMSTKNSWCRecord
from .programdb.nswc.table import RAMSTKNSWCTable
from .programdb.opload.record import RAMSTKOpLoadRecord
from .programdb.opload.table import RAMSTKOpLoadTable
from .programdb.opstress.record import RAMSTKOpStressRecord
from .programdb.opstress.table import RAMSTKOpStressTable
from .programdb.pof.view import RAMSTKPoFView
from .programdb.program_info.record import RAMSTKProgramInfoRecord
from .programdb.program_info.table import RAMSTKProgramInfoTable
from .programdb.program_status.record import RAMSTKProgramStatusRecord
from .programdb.program_status.table import RAMSTKProgramStatusTable
from .programdb.reliability.record import RAMSTKReliabilityRecord
from .programdb.reliability.table import RAMSTKReliabilityTable
from .programdb.requirement.record import RAMSTKRequirementRecord
from .programdb.requirement.table import RAMSTKRequirementTable
from .programdb.revision.record import RAMSTKRevisionRecord
from .programdb.revision.table import RAMSTKRevisionTable
from .programdb.similar_item.record import RAMSTKSimilarItemRecord
from .programdb.similar_item.table import RAMSTKSimilarItemTable
from .programdb.stakeholder.record import RAMSTKStakeholderRecord
from .programdb.stakeholder.table import RAMSTKStakeholderTable
from .programdb.test_method.record import RAMSTKTestMethodRecord
from .programdb.test_method.table import RAMSTKTestMethodTable
from .programdb.usage_profile.view import RAMSTKUsageProfileView
from .programdb.validation.record import RAMSTKValidationRecord
from .programdb.validation.table import RAMSTKValidationTable



# pylint: disable=non-parent-init-called
# -*- coding: utf-8 -*-
#
#       ramstk.gui.gtk.ramstk.Entry.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2017 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTK Entry Module."""

# Standard Library Imports
from datetime import datetime
from typing import Any, Callable

# RAMSTK Package Imports
from ramstk.views.gtk3 import Gtk, Pango

# RAMSTK Local Imports
from .widget import RAMSTKWidget


class RAMSTKEntry(Gtk.Entry, RAMSTKWidget):
    """The RAMSTK Entry class."""

    # Define private scalar class attributes.
    _default_height = 25
    _default_width = 200

    def __init__(self) -> None:
        """Create a RAMSTK Entry widget."""
        RAMSTKWidget.__init__(self)

        self.show()

    def do_get_text(self):
        """Wrap the Gtk.Entry().get_text() method for consistent calls.

        :return: the text in the RAMSTKEntry().
        :rtype: str
        """
        return self.get_text()

    def do_set_properties(self, **kwargs: Any) -> None:
        """Set the properties of the RAMSTK Entry.

        Accepts the following keyword arguments:

            * *width* -- width of the Gtk.Entry() widget.
                Default is 200.
            * *height* -- height of the Gtk.Entry() widget.
                Default is 25.
            * *editable* -- boolean indicating whether Gtk.Entry()
                should be editable.  Defaults to True.
            * *bold* -- boolean indicating whether text should be bold.
                Defaults to False.
            * *color* -- the hexidecimal color to set the background when
                the Gtk.Entry() is not editable.  Default is #BBDDFF (light
                blue).
            * *tooltip* -- the tooltip, if any, for the entry.
                Default is an empty string.
        :return: None
        :rtype: None
        """
        super().do_set_properties(**kwargs)

        _bold = kwargs.get("bold", False)
        _editable = kwargs.get("editable", True)

        self.set_property("editable", _editable)

        if _bold:
            self.modify_font(Pango.FontDescription("bold"))

    def do_update(self, value: Any, signal: str = "") -> None:
        """Update the RAMSTK Entry with a new value.

        :param value: the information to update the RAMSTKEntry() to
            display.
        :keyword signal: the name of the signal whose handler ID the
            RAMSTKEntry() needs to block.
        :return: None
        :rtype: None
        """
        try:
            _value = datetime.strftime(value, "%Y-%m-%d")
        except TypeError:
            _value = str(value)

        # Sometimes there is no handler ID for the RAMSTKEntry().  This
        # usually happens when the widget is being used to display results
        # and has no associated callback method.
        try:
            _handler_id = self.dic_handler_id[signal]

            self.handler_block(_handler_id)
            self.set_text(_value)
            self.handler_unblock(_handler_id)
        except KeyError:
            self.set_text(_value)


class RAMSTKTextView(Gtk.TextView, RAMSTKWidget):
    """The RAMSTK TextView class."""

    # Define private class scalar attributes.
    _default_height = 100
    _default_width = 200

    def __init__(self, txvbuffer: Gtk.TextBuffer) -> None:
        """Create RAMSTK TextView() widgets.

        Returns a Gtk.TextView() embedded in a Gtk.ScrolledWindow().

        :param txvbuffer: the Gtk.TextBuffer() to associate with the
            RAMSTKTextView().
        :return: None
        :rtype: None
        """
        RAMSTKWidget.__init__(self)

        self.set_buffer(txvbuffer)
        self.set_wrap_mode(Gtk.WrapMode.WORD)

        self.scrollwindow = Gtk.ScrolledWindow()
        self.scrollwindow.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.scrollwindow.add_with_viewport(self)

        self.tag_bold = txvbuffer.create_tag("bold", weight=Pango.Weight.BOLD)

    # pylint: disable=arguments-differ
    def connect(
        self,
        signal: str,
        callback: Callable,
        index: int,
        message: str,
    ) -> int:
        """Connect textview's buffer to a callback method or function.

        :param signal:
        :param callback:
        :param index:
        :param message:
        """
        _buffer = self.do_get_buffer()
        return _buffer.connect(signal, callback, index, message, self)

    def do_get_buffer(self) -> Gtk.TextBuffer:
        """Return the Gtk.TextBuffer() emedded in the RAMSTK TextView.

        :return: buffer; the embedded Gtk.TextBuffer()
        :rtype: :class:`Gtk.TextBuffer`
        """
        return self.get_buffer()

    def do_get_text(self) -> Any:
        """Retrieve the text from the embedded Gtk.TextBuffer().

        :return: text; the text in the Gtk.TextBuffer().
        :rtype: str
        """
        _buffer = self.do_get_buffer()

        return _buffer.get_text(*_buffer.get_bounds(), True)

    def do_set_properties(self, **kwargs: Any) -> None:
        """Set the properties of the RAMSTK TextView.

        Accepts the following keyword arguments:

            * *height* -- height of the Gtk.TextView() widget.  Default is 25.
            * *tooltip* -- the tooltip, if any, for the entry.  Default is an
                empty string.
            * *width* -- width of the Gtk.TextView() widget.  Default is 200.
        :return: None
        :rtype: None
        """
        super().do_set_properties(**kwargs)

        self.scrollwindow.set_property("height-request", self.height)
        self.scrollwindow.set_property("width-request", self.width)
        self.scrollwindow.set_shadow_type(Gtk.ShadowType.ETCHED_OUT)

    def do_update(self, value: str, signal: str = "") -> None:
        """Update the RAMSTK TextView with a new value.

        :param value: the information to update the RAMSTKTextView() to
            display.
        :keyword str signal: the name of the signal whose handler ID the
            RAMSTKTextView() needs to block.
        :return: None
        :rtype: None
        """
        _buffer = self.do_get_buffer()

        # Sometimes there is no handler ID for the RAMSTKTextView().  This
        # usually happens when the widget is being used to display results
        # and has no associated callback method.
        try:
            _handler_id = self.dic_handler_id[signal]

            _buffer.handler_block(_handler_id)
            _buffer.set_text(str(value))
            _buffer.handler_unblock(_handler_id)
        except KeyError:
            _buffer.set_text(str(value))



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.fmea.panel.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""GTK3 FMEA Panels."""

# Standard Library Imports
from typing import Any, Dict, List, Tuple, Union

# Third Party Imports
import treelib
from pubsub import pub

# RAMSTK Package Imports
from ramstk.models import RAMSTKCauseRecord, RAMSTKMechanismRecord
from ramstk.views.gtk3 import GdkPixbuf, Gtk, _
from ramstk.views.gtk3.widgets import (
    RAMSTKCheckButton,
    RAMSTKFixedPanel,
    RAMSTKLabel,
    RAMSTKTextView,
    RAMSTKTreePanel,
)


class FMEAMethodPanel(RAMSTKFixedPanel):
    """Panel to display FMEA criticality methods."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _record_field = "mode_id"
    _select_msg = "succeed_retrieve_fmea"
    _tag = "fmea"
    _title = _("FMEA Risk Analysis Method")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self):
        """Initialize an instance of the FMEA methods panel."""
        super().__init__()

        # Initialize widgets.
        self.chkCriticality: RAMSTKCheckButton = RAMSTKCheckButton(
            label=_("Calculate Criticality")
        )
        self.chkRPN: RAMSTKCheckButton = RAMSTKCheckButton(label=_("Calculate RPNs"))
        self.txtItemCriticality: RAMSTKTextView = RAMSTKTextView(Gtk.TextBuffer())

        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.
        self._on_edit_message = f"wvw_editing_{self._tag}"

        # Initialize public dictionary attributes.
        self.dic_attribute_widget_map = {
            "type_id": [
                27,
                self.chkCriticality,
                "toggled",
                super().on_toggled,
                self._on_edit_message,
                1,
                {
                    "tooltip": _(
                        "Select this option to calculate the MIL-STD-1629, Task 102 "
                        "criticality analysis."
                    ),
                },
                _("Calculate Criticality"),
                "gint",
            ],
            "rpn": [
                27,
                self.chkRPN,
                "toggled",
                super().on_toggled,
                self._on_edit_message,
                1,
                {
                    "tooltip": _(
                        "Select this option to calculate the Risk Priority Number "
                        "(RPN)."
                    ),
                },
                _("Calculate RPN"),
                "gint",
            ],
            "item_criticality": [
                28,
                self.txtItemCriticality,
                "changed",
                super().on_changed_entry,
                "",
                "",
                {
                    "bold": True,
                    "editable": False,
                    "height": 125,
                    "tooltip": _(
                        "Displays the MIL-STD-1629A, Task 102 item criticality for "
                        "the selected hardware item."
                    ),
                },
                _("Item Criticality:"),
                "gchararray",
            ],
        }

        # Initialize public list attributes.

        # Initialize public scalar attributes.

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        # Move the item criticality RAMSTKTextView() below it's label.
        _fixed: Gtk.Fixed = self.get_children()[0].get_children()[0].get_child()
        _label: RAMSTKLabel = _fixed.get_children()[-2]
        _x_pos: int = _fixed.child_get_property(_label, "x")
        _y_pos: int = _fixed.child_get_property(_label, "y") + 25
        _fixed.move(self.txtItemCriticality.scrollwindow, _x_pos, _y_pos)

        # Subscribe to PyPubSub messages.
        pub.subscribe(
            self._do_load_item_criticality,
            "succeed_calculate_mode_criticality",
        )

    def _do_load_item_criticality(self, item_criticality: Dict[str, float]) -> None:
        """Update the item criticality RAMSTKTextView() with the results.

        :param item_criticality: the item criticality for the selected
            hardware item.
        :return: None
        :rtype: None
        """
        _item_criticality = ""
        for _key, _value in item_criticality.items():
            _item_criticality = _item_criticality + _key + ": " + str(_value) + "\n"

        self.txtItemCriticality.do_update(_item_criticality, "changed")


class FMEATreePanel(RAMSTKTreePanel):
    """Panel to display FMEA analysis."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _select_msg = "succeed_retrieve_modes"
    _tag = "fmea"
    _title = _("Failure Mode and Effects Analysis")

    # Define public dictionary class attributes.

    # Define public dictionary list attributes.

    # Define public dictionary scalar attributes.

    # Define private dictionary class attributes.

    def __init__(self):
        """Initialize an instance of the FMEA analysis panel."""
        super().__init__()

        # Initialize private dictionary attributes.
        self._dic_mission_phases: Dict[str, List[str]] = {"": [""]}
        self.tvwTreeView.dic_row_loader = {
            "mode": self.__do_load_mode,
            "mechanism": self.__do_load_mechanism,
            "cause": self.__do_load_cause,
            "control": self.__do_load_control,
            "action": self.__do_load_action,
        }
        self._dic_visible_mask: Dict[str, List[str]] = {
            "mode": [
                "False",
                "False",
                "True",
                "False",
                "False",
                "False",
                "False",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "True",
                "True",
                "False",
                "True",
                "False",
            ],
            "mechanism": [
                "False",
                "False",
                "False",
                "True",
                "False",
                "False",
                "False",
                "True",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "True",
                "True",
                "True",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "True",
                "True",
                "True",
                "False",
                "False",
                "True",
                "False",
                "False",
            ],
            "cause": [
                "False",
                "False",
                "False",
                "False",
                "True",
                "False",
                "False",
                "True",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "True",
                "True",
                "True",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "True",
                "True",
                "True",
                "False",
                "False",
                "False",
                "False",
                "False",
            ],
            "control": [
                "False",
                "False",
                "False",
                "False",
                "False",
                "True",
                "False",
                "True",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
            ],
            "action": [
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "True",
                "True",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "True",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
                "False",
            ],
        }

        # Initialize private list attributes.
        self._lst_missions: List[str] = [""]

        # Initialize private scalar attributes.
        self._on_edit_message: str = f"wvw_editing_{self._tag}"

        # Initialize public dictionary attributes.
        self.dic_attribute_widget_map: Dict[str, List[Any]] = {
            "revision_id": [
                0,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Revision ID"),
                "gint",
            ],
            "hardware_id": [
                1,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Hardware ID"),
                "gint",
            ],
            "mode_id": [
                2,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Mode ID"),
                "gint",
            ],
            "mechanism_id": [
                3,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Mechanism ID"),
                "gint",
            ],
            "cause_id": [
                4,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Cause ID"),
                "gint",
            ],
            "control_id": [
                5,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Control ID"),
                "gint",
            ],
            "action_id": [
                6,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Action ID"),
                "gint",
            ],
            "description": [
                7,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Description"),
                "gchararray",
            ],
            "mission": [
                8,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Applicable Mission"),
                "gchararray",
            ],
            "mission_phase": [
                9,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_toggled,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Applicable Mission Phase"),
                "gchararray",
            ],
            "effect_local": [
                10,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Local Effect"),
                "gchararray",
            ],
            "effect_next": [
                11,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Next Effect"),
                "gchararray",
            ],
            "effect_end": [
                12,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("End Effect"),
                "gchararray",
            ],
            "detection_method": [
                13,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Detection Method"),
                "gchararray",
            ],
            "other_indications": [
                14,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Other Indications"),
                "gchararray",
            ],
            "isolation_method": [
                15,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Isolation Method"),
                "gchararray",
            ],
            "design_provisions": [
                16,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Design Provisions"),
                "gchararray",
            ],
            "operator_actions": [
                17,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Operator Actions"),
                "gchararray",
            ],
            "severity_class": [
                18,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Safety Severity"),
                "gchararray",
            ],
            "hazard_rate_source": [
                19,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Hazard Rate Data Source"),
                "gchararray",
            ],
            "mode_probability": [
                20,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Failure Probability"),
                "gchararray",
            ],
            "effect_probability": [
                21,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                1.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Effect Probability (beta)"),
                "gfloat",
            ],
            "mode_ratio": [
                22,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Mode Ratio (alpha)"),
                "gfloat",
            ],
            "mode_hazard_rate": [
                23,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Mode Hazard Rate"),
                "gfloat",
            ],
            "mode_op_time": [
                24,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Mode Operating Time"),
                "gfloat",
            ],
            "mode_criticality": [
                25,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Mode Criticality (Cm)"),
                "gfloat",
            ],
            "type_id": [
                26,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Control Type"),
                "gchararray",
            ],
            "rpn_severity": [
                27,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("RPN Severity"),
                "gint",
            ],
            "rpn_occurrence": [
                28,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("RPN Occurrence"),
                "gint",
            ],
            "rpn_detection": [
                29,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("RPN Detection"),
                "gint",
            ],
            "rpn": [
                30,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                1,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("RPN"),
                "gint",
            ],
            "action_category": [
                31,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Action Category"),
                "gchararray",
            ],
            "action_owner": [
                32,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Action Owner"),
                "gchararray",
            ],
            "action_due_date": [
                33,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Action Due Date"),
                "gchararray",
            ],
            "action_status": [
                34,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Action Status"),
                "gchararray",
            ],
            "action_taken": [
                35,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Action Taken"),
                "gchararray",
            ],
            "action_approved": [
                36,
                Gtk.CellRendererToggle(),
                "toggled",
                super().on_cell_toggled,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Action Approved"),
                "gint",
            ],
            "action_approve_date": [
                37,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Approval Date"),
                "gchararray",
            ],
            "action_closed": [
                38,
                Gtk.CellRendererToggle(),
                "toggled",
                super().on_cell_toggled,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Action Closed"),
                "gint",
            ],
            "action_close_date": [
                39,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Action Closure Date"),
                "gchararray",
            ],
            "rpn_severity_new": [
                40,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("New RPN Severity"),
                "gint",
            ],
            "rpn_occurrence_new": [
                41,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("New RPN Occurence"),
                "gint",
            ],
            "rpn_detection_new": [
                42,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("New RPN Detection"),
                "gint",
            ],
            "rpn_new": [
                43,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                1,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("New RPN"),
                "gint",
            ],
            "critical_item": [
                44,
                Gtk.CellRendererToggle(),
                "toggled",
                super().on_cell_toggled,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Critical Item"),
                "gint",
            ],
            "single_point": [
                45,
                Gtk.CellRendererToggle(),
                "toggled",
                super().on_cell_toggled,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Single Point"),
                "gint",
            ],
            "pof_include": [
                46,
                Gtk.CellRendererToggle(),
                "toggled",
                super().on_cell_toggled,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Include in PoF"),
                "gint",
            ],
            "remarks": [
                47,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Remarks"),
                "gchararray",
            ],
        }
        self.dic_icons: Dict[str, str] = {}

        # Initialize public list attributes.
        self.lst_action_category: List[str] = []
        self.lst_action_status: List[str] = []
        self.lst_control_types: List[str] = []
        self.lst_rpn_detection: List[str] = []
        self.lst_rpn_occurrence: List[str] = []
        self.lst_rpn_severity: List[str] = []
        self.lst_mode_probability: List[str] = []
        self.lst_severity_class: List[str] = []
        self.lst_users: List[str] = []

        # Initialize public scalar attributes.

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        self.tvwTreeView.set_tooltip_text(
            _(
                "Displays the (Design) Failure Mode and Effects "
                "(and Criticality) Analysis [(D)FME(C)A] for the "
                "currently selected Hardware item."
            )
        )

        # Subscribe to PyPubSub messages.
        pub.subscribe(super().do_load_panel, "succeed_retrieve_hardware_fmea")
        pub.subscribe(super().do_load_panel, "succeed_calculate_rpn")
        pub.subscribe(super().do_load_panel, "succeed_delete_action")
        pub.subscribe(super().do_load_panel, "succeed_delete_cause")
        pub.subscribe(super().do_load_panel, "succeed_delete_control")
        pub.subscribe(super().do_load_panel, "succeed_delete_mechanism")
        pub.subscribe(super().do_load_panel, "succeed_delete_mode")
        pub.subscribe(super().do_load_panel, "succeed_insert_action")
        pub.subscribe(super().do_load_panel, "succeed_insert_cause")
        pub.subscribe(super().do_load_panel, "succeed_insert_control")
        pub.subscribe(super().do_load_panel, "succeed_insert_mechanism")
        pub.subscribe(super().do_load_panel, "succeed_insert_mode")

        pub.subscribe(self.__do_load_missions, "succeed_retrieve_usage_profile")

    def do_load_comboboxes(self) -> None:
        """Load the Gtk.CellRendererCombo()s.

        :return: None
        :rtype: None
        """
        self.__do_load_action_category()
        self.__do_load_action_status()
        self.__do_load_control_type()
        self.__do_load_mode_probability()
        self.__do_load_rpn_detection()
        self.__do_load_rpn_occurrence()
        self.__do_load_rpn_severity()
        self.__do_load_severity_class()
        self.__do_load_users()

        _cell = self.tvwTreeView.get_column(
            self.tvwTreeView.position["mission"]
        ).get_cells()
        _cell[0].connect("edited", self._on_mission_change)

    def _on_mission_change(
        self, __combo: Gtk.CellRendererCombo, path: str, new_text: str
    ) -> None:
        """Load the mission phases whenever the mission combo is changed.

        :param __combo: the mission list Gtk.CellRendererCombo().  Unused in
            this method.
        :param path: the path identifying the edited cell.
        :param new_text: the new text (mission description).
        :return: None
        :rtype: None
        """
        self.__do_load_mission_phases(new_text)

        _model = self.tvwTreeView.get_model()
        _model[path][self.tvwTreeView.position["mission_phase"]] = ""

    def _on_row_change(self, selection: Gtk.TreeSelection) -> None:
        """Handle events for the FMEA Work View RAMSTKTreeView().

        This method is called whenever a RAMSTKTreeView() row is activated.

        :param selection: the current Gtk.TreeViewSelection() in the
            FMEA RAMSTKTreView().
        :type selection: :class:`Gtk.TreeViewSelection`
        :return: None
        :rtype: None
        """
        _model, _row = selection.get_selected()

        if _row is not None:
            if _model.get_value(_row, 3) == 0:
                _level = "mode"
            elif _model.get_value(_row, 4) == 0:
                _level = "mechanism"
            elif _model.get_value(_row, 5) == 0:
                _level = "cause"
            elif _model.get_value(_row, 6) == 0:
                _level = "control"
            else:
                _level = "action"

            self.tvwTreeView.visible = self._dic_visible_mask[_level]
            self.tvwTreeView.do_set_visible_columns()

            _mission = _model.get_value(_row, 8)
            self.__do_load_mission_phases(_mission)

    def __do_get_rpn_names(
        self, entity: Union[RAMSTKCauseRecord, RAMSTKMechanismRecord]
    ) -> Tuple[str, str, str, str]:
        """Retrieve the RPN category for the selected mechanism or cause.

        :param entity: the RAMSTKMechanism or RAMSTKCause object to be read.
        :return: (_occurrence, _detection, _occurrence_new, _detection_new)
        :rtype: tuple
        """
        _occurrence = str(self.lst_occurrence[entity.rpn_occurrence])
        _detection = str(self.lst_detection[entity.rpn_detection])
        _occurrence_new = str(self.lst_occurrence[entity.rpn_occurrence_new])
        _detection_new = str(self.lst_detection[entity.rpn_detection_new])

        return _occurrence, _detection, _occurrence_new, _detection_new

    def __do_load_action(self, node: treelib.Node, row: Gtk.TreeIter) -> Gtk.TreeIter:
        """Load an action record into the RAMSTKTreeView().

        :param node: the treelib Node() with the action data to load.
        :param row: the parent row of the action to load into the FMEA form.
        :return: _new_row; the row that was just populated with action data.
        :rtype: :class:`Gtk.TreeIter`
        """
        _new_row = None
        _date_format = "%Y-%m-%d"

        [[__, _entity]] = node.data.items()  # pylint: disable=unused-variable

        _model = self.tvwTreeView.get_model()

        # noinspection PyArgumentList
        _icon = GdkPixbuf.Pixbuf.new_from_file_at_size(self.dic_icons["action"], 22, 22)

        _attributes = [
            _entity.revision_id,
            _entity.hardware_id,
            _entity.mode_id,
            _entity.mechanism_id,
            _entity.cause_id,
            _entity.control_id,
            _entity.action_id,
            _entity.action_recommended,
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            "",
            "",
            "",
            "",
            0,
            _entity.action_category,
            _entity.action_owner,
            _entity.action_due_date.strftime(_date_format),
            _entity.action_status,
            _entity.action_taken,
            _entity.action_approved,
            _entity.action_approve_date.strftime(_date_format),
            _entity.action_closed,
            _entity.action_close_date.strftime(_date_format),
            "",
            "",
            "",
            0,
            0,
            0,
            0,
            "",
            _icon,
        ]

        try:
            _new_row = _model.append(row, _attributes)
        except (AttributeError, TypeError, ValueError):
            _new_row = None
            _message = _(
                "An error occurred when loading failure cause action {0:s} "
                "in the FMEA.  This might indicate it was missing it's data "
                "package, some of the data in the package was missing, or "
                "some of the data was the wrong type.  Row data was: "
                "{1}"
            ).format(str(node.identifier), _attributes)
            pub.sendMessage(
                "do_log_warning_msg", logger_name="WARNING", message=_message
            )

        return _new_row

    def __do_load_action_category(self) -> None:
        """Load the action category Gtk.CellRendererCombo().

        :return: None
        :rtype: None
        """
        self.tvwTreeView.do_load_combo_cell(
            self.tvwTreeView.position["action_category"], self.lst_action_category
        )

    def __do_load_action_status(self) -> None:
        """Load the action status Gtk.CellRendererCombo().

        :return: None
        :rtype: None
        """
        self.tvwTreeView.do_load_combo_cell(
            self.tvwTreeView.position["action_status"], self.lst_action_status
        )

    def __do_load_cause(self, node: treelib.Node, row: Gtk.TreeIter) -> Gtk.TreeIter:
        """Load a failure cause record into the RAMSTKTreeView().

        :param node: the treelib Node() with the cause data to load.
        :type node: :class:`treelib.Node`
        :param row: the parent row of the cause to load into the FMEA form.
        :type row: :class:`Gtk.TreeIter`
        :return: _new_row; the row that was just populated with cause data.
        :rtype: :class:`Gtk.TreeIter`
        """
        _new_row = None

        [[__, _entity]] = node.data.items()  # pylint: disable=unused-variable

        _model = self.tvwTreeView.get_model()

        (
            _occurrence,
            _detection,
            _occurrence_new,
            _detection_new,
        ) = self.__do_get_rpn_names(_entity)

        # noinspection PyArgumentList
        _icon = GdkPixbuf.Pixbuf.new_from_file_at_size(self.dic_icons["cause"], 22, 22)

        _attributes = [
            _entity.revision_id,
            _entity.hardware_id,
            _entity.mode_id,
            _entity.mechanism_id,
            _entity.cause_id,
            0,
            0,
            _entity.description,
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            "",
            "",
            _occurrence,
            _detection,
            _entity.rpn,
            "",
            "",
            "",
            "",
            "",
            0,
            "",
            0,
            "",
            "",
            _occurrence_new,
            _detection_new,
            _entity.rpn_new,
            0,
            0,
            0,
            "",
            _icon,
        ]

        try:
            _new_row = _model.append(row, _attributes)
        except (AttributeError, TypeError, ValueError):
            _new_row = None
            _message = _(
                f"An error occurred when loading failure cause {node.identifier} in "
                f"the FMEA.  This might indicate it was missing it's data package, "
                f"some of the data in the package was missing, or some of the "
                f"data was the wrong type.  Row data was: {_attributes}"
            )
            pub.sendMessage(
                "do_log_warning_msg", logger_name="WARNING", message=_message
            )

        return _new_row

    def __do_load_control(self, node: treelib.Node, row: Gtk.TreeIter) -> Gtk.TreeIter:
        """Load a control record into the RAMSTKTreeView().

        :param node: the treelib Node() with the control data to load.
        :type node: :class:`treelib.Node`
        :param row: the parent row of the control to load into the FMEA form.
        :type row: :class:`Gtk.TreeIter`
        :return: _new_row; the row that was just populated with control data.
        :rtype: :class:`Gtk.TreeIter`
        """
        _new_row = None

        [[__, _entity]] = node.data.items()  # pylint: disable=unused-variable

        _model = self.tvwTreeView.get_model()

        # noinspection PyArgumentList
        _icon = GdkPixbuf.Pixbuf.new_from_file_at_size(
            self.dic_icons["control"], 22, 22
        )

        _attributes = [
            _entity.revision_id,
            _entity.hardware_id,
            _entity.mode_id,
            _entity.mechanism_id,
            _entity.cause_id,
            _entity.control_id,
            0,
            _entity.description,
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            _entity.type_id,
            "",
            "",
            "",
            0,
            "",
            "",
            "",
            "",
            "",
            0,
            "",
            0,
            "",
            "",
            "",
            "",
            0,
            0,
            0,
            0,
            "",
            _icon,
        ]

        try:
            _new_row = _model.append(row, _attributes)
        except (AttributeError, TypeError, ValueError):
            _new_row = None
            _message = _(
                f"An error occurred when loading failure cause control "
                f"{node.identifier} in the FMEA.  This might indicate it was missing "
                f"it's data package, some of the data in the package was missing, or "
                f"some of the data was the wrong type.  Row data was: {_attributes}"
            )
            pub.sendMessage(
                "do_log_warning_msg", logger_name="WARNING", message=_message
            )

        return _new_row

    def __do_load_control_type(self) -> None:
        """Load the control type Gtk.CellRendererCombo().

        :return: None
        :rtype: None
        """
        self.tvwTreeView.do_load_combo_cell(
            self.tvwTreeView.position["type_id"], self.lst_control_types
        )

    def __do_load_mechanism(
        self, node: treelib.Node, row: Gtk.TreeIter
    ) -> Gtk.TreeIter:
        """Load a failure mechanism record into the RAMSTKTreeView().

        :param node: the treelib Node() with the mechanism data to load.
        :type node: :class:`treelib.Node`
        :param row: the parent row of the mechanism to load into the FMEA form.
        :type row: :class:`Gtk.TreeIter`
        :return: _new_row; the row that was just populated with mechanism data.
        :rtype: :class:`Gtk.TreeIter`
        """
        _new_row = None

        [[__, _entity]] = node.data.items()  # pylint: disable=unused-variable

        _model = self.tvwTreeView.get_model()

        (
            _occurrence,
            _detection,
            _occurrence_new,
            _detection_new,
        ) = self.__do_get_rpn_names(_entity)

        # noinspection PyArgumentList
        _icon = GdkPixbuf.Pixbuf.new_from_file_at_size(
            self.dic_icons["mechanism"], 22, 22
        )

        _attributes = [
            _entity.revision_id,
            _entity.hardware_id,
            _entity.mode_id,
            _entity.mechanism_id,
            0,
            0,
            0,
            _entity.description,
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            "",
            "",
            _occurrence,
            _detection,
            _entity.rpn,
            "",
            "",
            "",
            "",
            "",
            0,
            "",
            0,
            "",
            "",
            _occurrence_new,
            _detection_new,
            _entity.rpn_new,
            0,
            0,
            _entity.pof_include,
            "",
            _icon,
        ]

        try:
            _new_row = _model.append(row, _attributes)
        except (AttributeError, TypeError, ValueError):
            _new_row = None
            _message = _(
                f"An error occurred when loading failure mechanism {node.identifier} "
                f"in the FMEA.  This might indicate it was missing it's data "
                f"package, some of the data in the package was missing, or "
                f"some of the data was the wrong type.  Row data was: {_attributes}"
            )
            pub.sendMessage(
                "do_log_warning_msg", logger_name="WARNING", message=_message
            )

        return _new_row

    # noinspection PyUnusedLocal
    # pylint: disable=unused-argument
    def __do_load_missions(
        self,
        tree: treelib.Tree = treelib.Tree(),
        node_id: Any = "",
        row: Gtk.TreeIter = None,
    ) -> None:
        """Load the mission and mission phase dicts.

        :param tree: the treelib usage profile treelib.Tree().
        :param node_id: unused in this function.  Required so this method
            compatible with other listeners for the
            'succeed_retrieve_usage_profile' message.
        :param row: unused in this function.  Required so this method
            compatible with other listeners for the
            'succeed_retrieve_usage_profile' message.
        :return: None
        :rtype: None
        """
        _model = self.tvwTreeView.get_cell_model(self.tvwTreeView.position["mission"])

        self._lst_missions = []
        _model.append([""])
        for _node in tree.children(tree.root):
            _lst_phases: List[str] = [""]

            _mission = _node.data["usage_profile"].get_attributes()["description"]
            _model.append([_mission])
            self._lst_missions.append(_mission)

            for _node2 in tree.children(_node.identifier):
                _mission_phase = _node2.data["usage_profile"].get_attributes()[
                    "description"
                ]
                _lst_phases.append(_mission_phase)
            self._dic_mission_phases[_mission] = _lst_phases

    def __do_load_mission_phases(self, mission: str) -> None:
        """Load the mission phase Gtk.CellRendererCombo().

        :param mission: the mission that was selected.
        :return: None
        :rtype: None
        """
        _model = self.tvwTreeView.get_cell_model(
            self.tvwTreeView.position["mission_Phase"]
        )
        _model.clear()
        _model.append([""])

        try:
            for _phase in self._dic_mission_phases[mission]:
                _model.append([_phase])
        except KeyError:
            pass

    def __do_load_mode(self, node: treelib.Node, row: Gtk.TreeIter) -> Gtk.TreeIter:
        """Load a failure mode record into the RAMSTKTreeView().

        :param node: the treelib Node() with the mode data to load.
        :param row: the parent row of the mode to load into the FMEA form.
        :return: _new_row; the row that was just populated with mode data.
        :rtype: :class:`Gtk.TreeIter`
        """
        _new_row = None

        [[__, _entity]] = node.data.items()  # pylint: disable=unused-variable

        _model = self.tvwTreeView.get_model()

        _severity = self.lst_severity[_entity.rpn_severity]
        _severity_new = self.lst_severity[_entity.rpn_severity_new]

        # noinspection PyArgumentList
        _icon = GdkPixbuf.Pixbuf.new_from_file_at_size(self.dic_icons["mode"], 22, 22)

        _attributes = [
            _entity.revision_id,
            _entity.hardware_id,
            _entity.mode_id,
            0,
            0,
            0,
            0,
            _entity.description,
            _entity.mission,
            _entity.mission_phase,
            _entity.effect_local,
            _entity.effect_next,
            _entity.effect_end,
            _entity.detection_method,
            _entity.other_indications,
            _entity.isolation_method,
            _entity.design_provisions,
            _entity.operator_actions,
            _entity.severity_class,
            _entity.hazard_rate_source,
            _entity.mode_probability,
            _entity.effect_probability,
            _entity.mode_ratio,
            _entity.mode_hazard_rate,
            _entity.mode_op_time,
            _entity.mode_criticality,
            "",
            _severity,
            "",
            "",
            0,
            "",
            "",
            "",
            "",
            "",
            0,
            "",
            0,
            "",
            _severity_new,
            "",
            "",
            0,
            _entity.critical_item,
            _entity.single_point,
            0,
            _entity.remarks,
            _icon,
        ]

        try:
            _new_row = _model.append(row, _attributes)
        except (AttributeError, TypeError, ValueError):
            _new_row = None
            _message = _(
                f"An error occurred when loading failure mode {node.identifier} in the "
                f"FMEA.  This might indicate it was missing it's data package, "
                f"some of the data in the package was missing, or some of the "
                f"data was the wrong type.  Row data was: {_attributes}"
            )
            pub.sendMessage(
                "do_log_warning_msg", logger_name="WARNING", message=_message
            )

        return _new_row

    def __do_load_mode_probability(self) -> None:
        """Load the mode probability Gtk.CellRendererCombo().

        :return: None
        :rtype: None
        """
        self.tvwTreeView.do_load_combo_cell(
            self.tvwTreeView.position["mode_probability"], self.lst_mode_probability
        )

    def __do_load_rpn_detection(self) -> None:
        """Load the RPN detection Gtk.CellRendererCombo().

        :return: None
        :rtype: None
        """
        for _key in ["rpn_detection", "rpn_detection_new"]:
            self.tvwTreeView.do_load_combo_cell(
                self.tvwTreeView.position[_key], self.lst_rpn_detection
            )

    def __do_load_rpn_occurrence(self) -> None:
        """Load the RPN occurrence Gtk.CellRendererCombo().

        :return: None
        :rtype: None
        """
        for _key in ["rpn_occurrence", "rpn_occurrence_new"]:
            self.tvwTreeView.do_load_combo_cell(
                self.tvwTreeView.position[_key], self.lst_rpn_occurrence
            )

    def __do_load_rpn_severity(self) -> None:
        """Load the RPN severity Gtk.CellRendererCombo().

        :return: None
        :rtype: None
        """
        for _key in ["rpn_severity", "rpn_severity_new"]:
            self.tvwTreeView.do_load_combo_cell(
                self.tvwTreeView.position[_key], self.lst_rpn_severity
            )

    def __do_load_severity_class(self) -> None:
        """Load the severity classification Gtk.CellRendererCombo().

        :return: None
        :rtype: None
        """
        self.tvwTreeView.do_load_combo_cell(
            self.tvwTreeView.position["severity_class"], self.lst_severity_class
        )

    def __do_load_users(self) -> None:
        """Load the RAMSTK users Gtk.CellRendererCombo().

        :return: None
        :rtype: None
        """
        self.tvwTreeView.do_load_combo_cell(
            self.tvwTreeView.position["action_owner"], self.lst_users
        )



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.assistants.export.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2020 Doyle "weibullguy" Rowland
"""Export Assistant Module."""

# Standard Library Imports
import os

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration
from ramstk.views.gtk3 import Gtk, _
from ramstk.views.gtk3.widgets import RAMSTKFileChooser, RAMSTKMessageDialog


class ExportProject(RAMSTKFileChooser):
    """Assistant to walk user through the process of exporting records."""

    RAMSTK_USER_CONFIGURATION = RAMSTKUserConfiguration()

    def __init__(
        self,
        __button: Gtk.ToolButton,
        configuration: RAMSTKUserConfiguration,
        parent: object,
    ) -> None:
        """Initialize an instance of the Export Assistant.

        :param __button: the Gtk.ToolButton() that launched an instance of this
            class.
        :param configuration: the RAMSTKUserConfiguration class instance.
        :param parent: the parent window to associate this dialog with.
        """
        super().__init__(_("RAMSTK Export"), parent)

        # Initialize private dict variables.

        # Initialize private list variables.

        # Initialize private scalar variables.
        self._parent = parent

        # Initialize public dict variables.

        # Initialize public list variables.

        # Initialize public scalar variables.
        self.RAMSTK_USER_CONFIGURATION = configuration

        self.set_current_folder(self.RAMSTK_USER_CONFIGURATION.RAMSTK_PROG_DIR)

        for _module in ["functions", "requirements", "hardwares", "validations"]:
            pub.sendMessage("request_get_{0}_tree".format(_module))

        self._do_select_file()

    def _do_quit(self) -> None:
        """Quit the RAMSTK Export Assistant.

        :return: None
        :rtype: None
        """
        self.destroy()

    def _do_select_file(self) -> None:
        """Select the input file to export data to.

        :return: None
        :rtype: None
        """
        _cansave = False
        (_filename, _extension) = self.do_run()

        if _filename is not None:
            _filetype = {
                ".csv": "csv",
                ".txt": "text",
                ".xls": "excel",
                ".xlsm": "excel",
                ".xlsx": "excel",
            }[
                _extension
            ]  # type: ignore

            if os.path.exists(_filename):
                _dialog = RAMSTKMessageDialog(self._parent)
                _dialog.do_set_message(
                    _("File {0} already exists.  Overwrite?").format(_filename)
                )
                _dialog.do_set_message_type("question")
                _response = _dialog.do_run()
                if _response == Gtk.ResponseType.YES:
                    _dialog.destroy()
                    os.remove(_filename)
                    _cansave = True
                else:
                    _dialog.destroy()
            else:
                _cansave = True

            if _cansave:
                pub.sendMessage(
                    "request_export_data",
                    file_type=_filetype,
                    file_name=_filename,
                )
            else:
                self._do_select_file()

        self._do_quit()



# Standard Library Imports
from typing import Any, Dict, List

# RAMSTK Package Imports
from ramstk.views.gtk3 import _ as _
from ramstk.views.gtk3.widgets import RAMSTKCheckButton as RAMSTKCheckButton
from ramstk.views.gtk3.widgets import RAMSTKEntry as RAMSTKEntry
from ramstk.views.gtk3.widgets import RAMSTKFixedPanel as RAMSTKFixedPanel

class OptionsPanel(RAMSTKFixedPanel):
    _select_msg: str
    _tag: str
    _title: str
    chkFunctions: RAMSTKCheckButton
    chkRequirements: RAMSTKCheckButton
    chkHardware: RAMSTKCheckButton
    chkValidation: RAMSTKCheckButton
    chkHazards: RAMSTKCheckButton
    chkStakeholder: RAMSTKCheckButton
    chkAllocation: RAMSTKCheckButton
    chkSimilarItem: RAMSTKCheckButton
    chkFMEA: RAMSTKCheckButton
    chkPoF: RAMSTKCheckButton
    txtSiteID: RAMSTKEntry
    txtSiteName: RAMSTKEntry
    txtProductKey: RAMSTKEntry
    txtExpireDate: RAMSTKEntry
    dic_attribute_widget_map: Dict[str, List[Any]]
    def __init__(self) -> None: ...



# pylint: disable=protected-access, non-parent-init-called
# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.ramstk.Button.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2017 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTK Button Module."""

# Standard Library Imports
from typing import Any, Union

# RAMSTK Package Imports
from ramstk.views.gtk3 import GdkPixbuf, Gtk, _

# RAMSTK Local Imports
from .widget import RAMSTKWidget


def do_make_buttonbox(
    view: Any, **kwargs: Any
) -> Union[Gtk.HButtonBox, Gtk.VButtonBox]:
    """Create the buttonbox for RAMSTK Views.

    This method creates the base buttonbox used by all RAMSTK Views.  Use a
    buttonbox for a RAMSTK View if there are only buttons to be added.

    :param view: the RAMSTKView() this buttonbox will be embedded in.
    :type view: :class:`ramstk.views.gtk3.widgets.view.RAMSTKBaseView`

    Accepts the following keyword arguments:
        * *callbacks* (list) -- list of callback functions to assign to each
            button in the buttonbox.
        * *height* (int) -- height of the Gtk.Button() widget.
            Default is 30.
        * *icons* (list) -- list of the icons to display on each button in the
            buttonbox.
        * *orientation* (str) -- the orientation of the buttonbox, either
            horizontal or vertical.  Default is vertical.
        * *tooltips* (list) -- list of tooltips for each button in the
            buttonbox.
        * *width* (int) -- width of the Gtk.Button() widget.
            Default is 200.

    :return: _buttonbox
    :rtype: :class:`Gtk.ButtonBox`
    """
    _callbacks = kwargs.get("callbacks", [])
    _height = kwargs.get("height", -1)
    _icons = kwargs.get("icons", [])
    _orientation = kwargs.get("orientation", "vertical")
    _tooltips = kwargs.get("tooltips", [])
    _width = kwargs.get("width", -1)

    if _orientation == "horizontal":
        _buttonbox = Gtk.HButtonBox()
    else:
        _buttonbox = Gtk.VButtonBox()

    _buttonbox.set_layout(Gtk.ButtonBoxStyle.START)

    for _idx, _icon in enumerate(_icons):
        _image = Gtk.Image()
        _icon = GdkPixbuf.Pixbuf.new_from_file_at_size(
            view._dic_icons[_icon], _height, _width
        )
        _image.set_from_pixbuf(_icon)

        _button = Gtk.Button()
        _button.set_image(_image)

        _button.props.width_request = _width
        _button.props.height_request = _height

        try:
            _button.set_tooltip_markup(_tooltips[_idx])
        except IndexError:
            _button.set_tooltip_markup("")

        try:
            _button.connect("clicked", _callbacks[_idx])
        except IndexError:
            _button.set_sensitive(False)

        _buttonbox.pack_start(_button, True, True, 0)

    return _buttonbox


class RAMSTKButton(Gtk.Button, RAMSTKWidget):
    """The RAMSTK Button class."""

    # Define private class scalar attributes.
    _default_height = 30
    _default_width = 200

    def __init__(self, label: str = "...") -> None:
        """Initialize an instance of the RAMSTKButton().

        :keyword str label: the text to display on the RAMSTKButton().
            Default is an ellipsis (...).
        :return: None
        :rtype: None
        """
        # noinspection PyCallByClass,PyTypeChecker
        RAMSTKWidget.__init__(self)

        self.set_label(label)
        self.show_all()

    def do_set_properties(self, **kwargs: Any) -> None:
        """Set the properties of the RAMSTKButton.

        Accepts the following keyword arguments:
            * *icon* (str) -- the icon to display on the button.  Default is
                None.
        :return: None
        :rtype: None
        """
        super().do_set_properties(**kwargs)
        self.set_property("always_show_image", True)

        _icon = kwargs.get("icon", None)

        if _icon is not None:
            self.set_label("")
            _icon = GdkPixbuf.Pixbuf.new_from_file_at_size(
                _icon, self.height, self.width
            )
            _image = Gtk.Image()
            _image.set_from_pixbuf(_icon)
            self.set_image(_image)


class RAMSTKCheckButton(Gtk.CheckButton, RAMSTKWidget):
    """The RAMSTK Check Button class."""

    # Define private class scalar attributes.
    _default_height = 40
    _default_width = 200

    def __init__(self, label: str = "") -> None:
        """Initialize an instance of the RAMSTK CheckButton.

        :keyword str label: the text to display with the Gtk.CheckButton().
            Default is an empty string.
        :return: None
        :rtype: None
        """
        # noinspection PyCallByClass,PyTypeChecker
        RAMSTKWidget.__init__(self)

        self.set_label(label)

    def do_set_properties(self, **kwargs: Any) -> None:
        """Set the properties of the RAMSTK button.

        Accepts the following keyword arguments:
            * *height* (int) -- height of the RAMSTKCheckButton() widget.
                Default is 40.
            * *tooltip* (str) -- the tooltip, if any, for the button.
                Default is a message to file a QA-type issue to have one added.
            * *width* (int) -- width of the RAMSTKCheckButton() widget.
                Default is 200.
        :return: None
        :rtype: None
        """
        super().do_set_properties(**kwargs)

        self.set_use_underline(True)

        self.get_child().set_use_markup(True)
        self.get_child().set_line_wrap(True)
        self.get_child().set_property("wrap", True)
        self.get_child().set_property("height-request", self.height)
        self.get_child().set_property("width-request", self.width)

    def do_update(self, value: int, signal: str = "") -> None:
        """Update the RAMSTK CheckButton with a new value.

        :param value: the information to update the RAMSTKCheckButton() to
            display.
        :keyword str signal: the name of the signal whose handler ID the
            RAMSTKCheckButton() needs to block.
        :return: None
        :rtype: None
        """
        _handler_id = self.dic_handler_id[signal]

        self.handler_block(_handler_id)
        self.set_active(int(value))
        self.handler_unblock(_handler_id)


class RAMSTKFileChooserButton(Gtk.FileChooserButton, RAMSTKWidget):
    """The RAMSTK File Chooser Button class."""

    def __init__(self, label: str = "...") -> None:
        """Initialize an instance of the RAMSTK FileChooserButton.

        :param label: the text for the button's label.
        :return: None
        :rtype: None
        """
        RAMSTKWidget.__init__(self)

        self.set_title(label)
        self.show_all()

    def do_set_properties(self, **kwargs: Any) -> None:
        """Set the properties of the RAMSTKFileChooserButton.

        Accepts the following keyword arguments:

            * *height* (int) -- height of the RAMSTKFileChooserButton() widget.
                Default is 30.
            * *select-action*  -- the Gtk.FileChooserAction to set for the button.
            * *tooltip* (str) -- the tooltip, if any, for the RAMSTKFileChooserButton().
                Default is a message to file a QA-type issue to have one added.
            * *width* (int) -- width of the RAMSTKFileChooserButton() widget.
                Default is 200.
        :return: None
        :rtype: None
        """
        super().do_set_properties(**kwargs)

        _select_action = kwargs.get("select-action", None)

        self.set_action(_select_action)


class RAMSTKOptionButton(Gtk.RadioButton, RAMSTKWidget):
    """The RAMSTK Option Button class."""

    def __init__(self, group: Gtk.RadioButton = None, label: str = _("")) -> None:
        """Initialize an instance of the RAMSTK OptionButton.

        :keyword group: the group the Gtk.RadioButton() belongs to, if any.
            Default is None.
        :type group: :class:`Gtk.RadioButton`
        :keyword str label: the text to place in the label on the
            Gtk.RadioButton().  Default is an empty string.
        :return: None
        :rtype: None
        """
        # noinspection PyCallByClass,PyTypeChecker
        RAMSTKWidget.__init__(self)

        self.set_group(group)
        self.set_label(label)


class RAMSTKSpinButton(Gtk.SpinButton, RAMSTKWidget):
    """The RAMSTK Spin Button class."""

    # Define private class scalar attributes.
    _default_height = 30
    _default_width = 200

    def __init__(self) -> None:
        """Initialize an instance of the RAMSTKSpinButton().

        :return: None
        :rtype: None
        """
        # noinspection PyCallByClass,PyTypeChecker
        RAMSTKWidget.__init__(self)

        self.show_all()

    def do_get_text(self) -> float:
        """Wrap the Gtk.SpinButton().get_value() method for consistent calls.

        :return: the float value of the text in the RAMSTKSpinButton().
        :rtype: float
        """
        return float(self.get_value())

    def do_set_properties(self, **kwargs: Any) -> None:
        """Set the properties of the RAMSTKSpinButton.

        Accepts the following keyword arguments:

            * *height* (int) -- height of the RAMSTKButton() widget.
                Default is 30.
            * *limits* (list) -- the list of values for the spin button
                Gtk.Adjustment().
            * *tooltip* (str) -- the tooltip, if any, for the button.
                Default is a message to file a QA-type issue to have one added.
            * *width* (int) -- width of the RAMSTKButton() widget.
                Default is 200.
        :return: None
        :rtype: None
        """
        super().do_set_properties(**kwargs)

        _limits = kwargs.get("limits", [0, 0, 100, 1, 0.1])
        _numeric = kwargs.get("numeric", True)
        _snap_to_ticks = kwargs.get("ticks", True)

        self.set_adjustment(
            Gtk.Adjustment(_limits[0], _limits[1], _limits[2], _limits[3], _limits[4])
        )
        self.set_numeric(_numeric)
        self.set_snap_to_ticks(_snap_to_ticks)
        self.set_update_policy(Gtk.SpinButtonUpdatePolicy.IF_VALID)

    def do_update(self, value: int, signal: str = "") -> None:
        """Update the RAMSTK Spin Button with a new value.

        :param value: the information to update the RAMSTKSpinButton() to
            display.
        :keyword str signal: the name of the signal whose handler ID the
            RAMSTKSpinButton() needs to block.
        :return: None
        :rtype: None
        """
        _handler_id = self.dic_handler_id[signal]

        self.handler_block(_handler_id)
        self.set_value(int(value))
        self.handler_unblock(_handler_id)



# pylint: disable=unused-import, missing-docstring
# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.options.view.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTK GTK3 Options Views."""

# RAMSTK Package Imports
from ramstk.views.gtk3 import Gtk, _
from ramstk.views.gtk3.widgets import RAMSTKDialog, RAMSTKPanel

# RAMSTK Local Imports
from . import OptionsPanel


class OptionsDialog(RAMSTKDialog):
    """Provide a GUI to set various RAMSTK configuration options.

    RAMSTK options are stored in the RAMSTK Common database and the RAMSTK
    Program database.  RAMSTK options are site-specific or program-specific and
    apply to all users.  Options should not be confused with user-specific
    configurations preferences which are stored in RAMSTK.conf in each user's
    $HOME/.config/RAMSTK directory and are applicable only to that specific
    user.  Configuration preferences are edited with the Preferences assistant.

    Attributes of the EditOptions are:
    """

    # Define private dict class attributes.

    def __init__(self, parent: object = None) -> None:
        """Initialize an instance of the Options assistant.

        :param parent: the parent window for this assistant.
        """
        super().__init__(_("RAMSTK Program Options Assistant"), dlgparent=parent)

        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.
        self._pnlPanel: RAMSTKPanel = OptionsPanel()

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.

        self.__make_ui()

        # Subscribe to PyPubSub messages.

    def _cancel(self, __button: Gtk.Button):
        """Destroy the assistant when the 'Cancel' button is pressed.

        :param __button: the Gtk.Button() that called this method.
        :type __button: :class:`Gtk.Button`
        """
        self.do_destroy()

    def __make_ui(self) -> None:
        """Build the user interface.

        :return: None
        :rtype: None
        """
        self.set_default_size(800, 500)

        self.vbox.pack_start(self._pnlPanel, True, True, 0)

        self.show_all()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ramstk',
 'ramstk.analyses',
 'ramstk.analyses.milhdbk217f',
 'ramstk.analyses.milhdbk217f.models',
 'ramstk.analyses.statistics',
 'ramstk.db',
 'ramstk.exim',
 'ramstk.models',
 'ramstk.models.commondb',
 'ramstk.models.commondb.site_info',
 'ramstk.models.programdb',
 'ramstk.models.programdb.action',
 'ramstk.models.programdb.allocation',
 'ramstk.models.programdb.cause',
 'ramstk.models.programdb.control',
 'ramstk.models.programdb.design_electric',
 'ramstk.models.programdb.design_mechanic',
 'ramstk.models.programdb.environment',
 'ramstk.models.programdb.failure_definition',
 'ramstk.models.programdb.fmea',
 'ramstk.models.programdb.function',
 'ramstk.models.programdb.hardware',
 'ramstk.models.programdb.hazard',
 'ramstk.models.programdb.mechanism',
 'ramstk.models.programdb.milhdbk217f',
 'ramstk.models.programdb.mission',
 'ramstk.models.programdb.mission_phase',
 'ramstk.models.programdb.mode',
 'ramstk.models.programdb.nswc',
 'ramstk.models.programdb.opload',
 'ramstk.models.programdb.opstress',
 'ramstk.models.programdb.pof',
 'ramstk.models.programdb.program_info',
 'ramstk.models.programdb.program_status',
 'ramstk.models.programdb.reliability',
 'ramstk.models.programdb.requirement',
 'ramstk.models.programdb.revision',
 'ramstk.models.programdb.similar_item',
 'ramstk.models.programdb.stakeholder',
 'ramstk.models.programdb.test_method',
 'ramstk.models.programdb.usage_profile',
 'ramstk.models.programdb.validation',
 'ramstk.views',
 'ramstk.views.gtk3',
 'ramstk.views.gtk3.allocation',
 'ramstk.views.gtk3.assistants',
 'ramstk.views.gtk3.books',
 'ramstk.views.gtk3.design_electric',
 'ramstk.views.gtk3.design_electric.components',
 'ramstk.views.gtk3.failure_definition',
 'ramstk.views.gtk3.fmea',
 'ramstk.views.gtk3.function',
 'ramstk.views.gtk3.hardware',
 'ramstk.views.gtk3.hazard_analysis',
 'ramstk.views.gtk3.milhdbk217f',
 'ramstk.views.gtk3.milhdbk217f.components',
 'ramstk.views.gtk3.options',
 'ramstk.views.gtk3.pof',
 'ramstk.views.gtk3.preferences',
 'ramstk.views.gtk3.program_status',
 'ramstk.views.gtk3.reliability',
 'ramstk.views.gtk3.requirement',
 'ramstk.views.gtk3.revision',
 'ramstk.views.gtk3.similar_item',
 'ramstk.views.gtk3.stakeholder',
 'ramstk.views.gtk3.usage_profile',
 'ramstk.views.gtk3.validation',
 'ramstk.views.gtk3.widgets']

package_data = \
{'': ['*']}

install_requires = \
['XlsxWriter>=3.0.1,<4.0.0',
 'lifelines>=0.26.3,<0.27.0',
 'matplotlib>=3.3.4,<4.0.0',
 'openpyxl>=3.0.6,<4.0.0',
 'pandas==1.1.5',
 'psycopg2>=2.8.6,<3.0.0',
 'pycairo>=1.20.0,<2.0.0',
 'pygobject>=3.38,<4.0',
 'pypubsub>=4.0.3,<5.0.0',
 'sortedcontainers>=2.3.0,<3.0.0',
 'sqlalchemy-utils>=0.37.8,<0.38.0',
 'sqlalchemy>=1.3.0,<2.0.0',
 'statsmodels>=0.12.2,<0.13.0',
 'sympy>=1.8.0,<2.0.0',
 'toml>=0.10.2,<0.11.0',
 'treelib>=1.5.3,<2.0.0',
 'xlrd>=2.0.1,<3.0.0',
 'xlwt>=1.3.0,<2.0.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['numpy>=1.19,<2.0',
                                                         'scipy>=1.5,<2.0'],
 ':python_version >= "3.7" and python_version < "4.0"': ['numpy>=1.21,<2.0',
                                                         'scipy>=1.6.0,<2.0.0']}

entry_points = \
{'console_scripts': ['ramstk = ramstk.__main__:the_one_ring']}

setup_kwargs = {
    'name': 'ramstk',
    'version': '0.15.0',
    'description': 'A tool to manage RAMS data and analyses.',
    'long_description': '# The RAMS ToolKit (RAMSTK)\n> A ToolKit for **R**eliability, **A**vailability, **M**aintainability, and\n> **S**afety (RAMS) analyses.\n\n<table>\n    <tr>\n        <th>Documentation</th>\n        <td>\n            <a href=\'https://ramstk.readthedocs.io/en/latest/?badge=latest\'><img src=\'https://readthedocs.org/projects/ramstk/badge/?version=latest\' alt=\'Documentation Status\' /></a>\n        </td>\n    </tr>\n    <tr>\n        <th>Tests</th>\n        <td>\n        <img alt="GitHub Workflow Status" src="https://img.shields.io/github/workflow/status/ReliaQualAssociates/ramstk/RAMSTK%20Test%20Suite?label=Build%20%26%20Test">\n        <a href="https://codecov.io/gh/ReliaQualAssociates/ramstk"><img src="https://codecov.io/gh/ReliaQualAssociates/ramstk/branch/master/graph/badge.svg?token=sFOa7EjZAg"/></a>\n        <a href=\'https://coveralls.io/github/ReliaQualAssociates/ramstk?branch=master\'><img src=\'https://coveralls.io/repos/github/ReliaQualAssociates/ramstk/badge.svg?branch=master\' alt=\'Coverage Status\' /></a>\n    </td>\n    </tr>\n    <tr>\n        <th>Quality</th>\n        <td>\n            <a href="https://www.codefactor.io/repository/github/reliaqualassociates/ramstk"><img src="https://www.codefactor.io/repository/github/reliaqualassociates/ramstk/badge" alt="CodeFactor" /></a>\n            <a href="https://www.deepcode.ai/app/gh/ReliaQualAssociates/ramstk/_/dashboard?utm_content=gh%2FReliaQualAssociates%2Framstk"><img alt="Deep Code" src="https://www.deepcode.ai/api/gh/badge?key=eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJwbGF0Zm9ybTEiOiJnaCIsIm93bmVyMSI6IlJlbGlhUXVhbEFzc29jaWF0ZXMiLCJyZXBvMSI6InJhbXN0ayIsImluY2x1ZGVMaW50IjpmYWxzZSwiYXV0aG9ySWQiOjI1MTA4LCJpYXQiOjE2MDkxMzcwNTl9.R5P6VLkyK1LK6Jc5PjJ8QrLRq6zNuVxnzdjZCJbH7_k"></a>\n            <img alt="Quality Gate" src="https://sonarcloud.io/api/project_badges/measure?project=ReliaQualAssociates_ramstk&metric=alert_status">\n        </td>\n    </tr>\n    <tr>\n        <th>Packages</th>\n        <td>\n            <img alt="GitHub release (latest SemVer including pre-releases)" src="https://img.shields.io/github/v/release/ReliaQualAssociates/ramstk?include_prereleases&label=GitHub%20Release">\n            <img alt="PyPI" src="https://img.shields.io/pypi/v/ramstk?label=PyPi%20Release">\n        </td>\n    </tr>\n</table>\n\n## 🚩 Table of Contents\n- [Features](#-features)\n- [Installing](#-installing)\n    - [Prerequisites](#prerequisites)\n    - [Download](#download)\n    - [Running the Tests](#running-the-tests)\n- [Usage](#-usage)\n- [Documentation](#documentation)\n- [Contributing](#-contributing)\n- [Authors](#-authors)\n- [License](#-license)\n- [Similar Products](#similar-products)\n\n## Disclaimer\n\nRAMSTK attempts to use [Semantic Versioning](https://semver.org/) 2.0.0.  Per\n4, major version 0 is for initial development and anything may change at\nany time.  That is certainly the case for RAMSTK!  Because RAMSTK is a one\ndeveloper show, there is no active develop branch at the moment.  This may\nchange after release of 1.0.0.  Until then, tagged releases can be used, but\nthe `latest` tag may not work and may not be backwards-compatible.  While major\nversion is at 0, breaking changes will be reflected in bumps to the minor\nversion number.  That is, version 0.15.0 is not compatible with version 0.14.0.\nAlso at this time, patch versions will not be released.  This will change after\nversion 1.0.0 is released.\n\n## 🎨&nbsp; Features\n\nRAMSTK is built on the concept of modules where a module is a collection of\n related information and/or analyses pertinent to system development.  The\n  modules currently included in RAMSTK are:\n\n* Function Module\n  - Functional decomposition\n  - Hazards analysis\n* Requirements Module\n  - Stakeholder input prioritization\n  - Requirement development\n  - Analysis of requirement for clarity, completeness, consistency, and verifiability\n* Hardware Module\n  - Reliability allocation\n      - Equal apportionment\n      - AGREE apportionment\n      - ARINC apportionment\n      - Feasibility of Objectives\n  - Hardware reliability predictions using various methods\n      - Similar items analysis\n      - MIL-HDBK-217F parts count\n      - MIL-HDBK-217F parts stress\n  - FMEA/FMECA\n      - RPN\n      - MIL-STD-1629A, Task 102 Criticality Analysis\n  - Physics of failure analysis\n* Validation Module\n  - Task description\n  - Task acceptance value(s)\n  - Task time\n  - Task cost\n  - Overall validation plan time/cost estimates\n\n## 💾&nbsp; Installing\n\nThese instructions will hopefully get RAMSTK up and running on your local\nmachine.  RAMSTK uses a Makefile to install/uninstall itself because there are\nvarious icon, data, and configuration files that also need to be installed\noutside site-packages.  Thus, only the actual RAMSTK application is available\nat PyPi and the initial installation must be done using the source asset at\nGitHub for the release you wish to install or cloning the RAMSTK repository if\nyou\'d like the latest code.\n\nRAMSTK uses [postgresql](https://www.postgresql.org/) for its database\n engine.  You\'ll need to have a user with read/write access to a postgresql\n  server to use RAMSTK.\n\n### Download and Install\n\nInstall any missing RAMSTK dependencies using pip, your package manager, and/or\nbuild from source.  Then download the \\<version> of RAMSTK source from GitHub\nyou wish to install.\n\n```shell\n$ wget https://github.com/ReliaQualAssociates/ramstk/archive/v<version>.tar.gz\n$ tar -xf v<version>.tar.gz\n$ cd ramstk-<version>\n$ make install\n```\n\nThe install target recognizes PREFIX=<non-default install path> so you can\n install RAMSTK in your $HOME or a virtual environment.  Since RAMSTK is\n  still a version 0 product, it\'s highly recommended that you install in a\n   virtual environment.\n\n```shell\n$ wget https://github.com/ReliaQualAssociates/ramstk/archive/v<version>.tar.gz\n$ tar -xf v<version>.tar.gz\n$ cd ramstk-<version>\n$ make PREFIX=$VIRTUAL_ENV install\n```\n\nWhen upgrading RAMSTK, you can simply:\n\n```shell\n$ pip install\n```\n\nThis will only install the latest RAMSTK version from PyPi and will leave\nconfiguration, data, and icon files untouched.  If you are using the latest\ncode from GitHub, you can also use the Makefile:\n\n```shell\n$ make install.dev\n```\n\n### Development Dependencies\n\nI use [poetry](https://github.com/python-poetry/poetry) to manage the\ndependencies for RAMSTK while I\'m developing.  Using the Makefile, install as\nfollows:\n\n```shell\n$ make depends\n```\n\nThis should get all the needed development and runtime requirements installed\nif they\'re not already.\n\n### Running the Tests\n\nTo run the entire test suite for RAMSTK after installing, simply execute:\n\n```shell\n$ make test\n```\n\nTo run the test suite with coverage, execute:\n\n```shell\n$ make coverage\n```\n\nTo run specific tests or groups of tests, use pytest:\n\n```shell\n$ pytest -m integration tests/modules/test_allocation.py\n$ pytest -m calculation tests/analyses/prediction\n```\n\n## 🔨&nbsp; Usage\n\nAfter installing RAMSTK, it can be launched from a terminal emulator:\n\n```\n$ ramstk\n```\n\nThis is a good option if you need to file an issue as the output should be\n included in your report.\n\nRAMSTK installs a *.desktop file and can be found where ever applications in\n the category Math or Science are listed.\n\n## Documentation\n\nDocumentation for RAMSTK can be found at [Read the Docs](https://ramstk.readthedocs.io/en/latest) You should check it out!\n\n## 💬&nbsp; Contributing\n\nPlease read [CONTRIBUTING.md](https://github.com/ReliaQualAssociates/ramstk/tree/develop/docs/CONTRIBUTING.md) for details on our code of conduct, and the process for submitting pull requests to us.\n\nAlso read [DEVELOPMENT_ENV.md](https://github.com/ReliaQualAssociates/ramstk/tree/develop/docs/DEVELOPMENT_ENV.md) for instructions on setting up a development environment to work on and test RAMSTK.\n\n## 🍞&nbsp; Authors\n\n* **Doyle \'weibullguy\' Rowland** - *Initial work* - [weibullguy](https://github.com/weibullguy)\n\n## 📜&nbsp; License\nThis project is licensed under the BSD-3-Clause License - see the [LICENSE](https://github.com/ReliaQualAssociates/ramstk/blob/develop/LICENSE) file for details.\n\nRAMSTK is also registered with the United States Copyright Office under\n registration number TXu 1-896-035.\n\n## Similar Products\n\nThe following are commercially available products that perform RAMS analyses\n.  We are not endorsing any of them; they are all fine products and may be a\n better fit for you or your organization depending on your needs and budget\n .  Obviously, we would prefer you use RAMSTK.\n\n* [PTC Windchill Quality](https://www.ptc.com/en/products/plm/capabilities/quality)\n* [ReliaSoft Synthesis](https://www.reliasoft.com/products)\n',
    'author': "Doyle 'weibullguy' Rowland",
    'author_email': 'doyle.rowland@reliaqual.com',
    'maintainer': "Doyle 'weibullguy' Rowland",
    'maintainer_email': 'doyle.rowland@reliaqual.com',
    'url': 'https://github.com/ReliaQualAssociates/ramstk',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



# pylint: disable=duplicate-code
# -*- coding: utf-8 -*-
#
#       ramstk.models.function.record.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Function Record Model."""

# Third Party Imports
from sqlalchemy import Column, Float, ForeignKey, Integer, String
from sqlalchemy.orm import relationship

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord


class RAMSTKFunctionRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    """Class to represent ramstk_function table in the RAMSTK Program database.

    This table shares a Many-to-One relationship with ramstk_revision. This table
    shares a One-to-Many relationship with ramstk_mode.
    """

    __defaults__ = {
        "availability_logistics": 1.0,
        "availability_mission": 1.0,
        "cost": 0.0,
        "function_code": "Function Code",
        "hazard_rate_logistics": 0.0,
        "hazard_rate_mission": 0.0,
        "level": 0,
        "mmt": 0.0,
        "mcmt": 0.0,
        "mpmt": 0.0,
        "mtbf_logistics": 0.0,
        "mtbf_mission": 0.0,
        "mttr": 0.0,
        "name": "Function Name",
        "parent_id": 0,
        "remarks": "",
        "safety_critical": 0,
        "total_mode_count": 0,
        "total_part_count": 0,
        "type_id": 0,
    }
    __tablename__ = "ramstk_function"
    __table_args__ = {"extend_existing": True}

    revision_id = Column(
        "fld_revision_id",
        Integer,
        ForeignKey("ramstk_revision.fld_revision_id", ondelete="CASCADE"),
        nullable=False,
    )
    function_id = Column(
        "fld_function_id",
        Integer,
        primary_key=True,
        autoincrement=True,
        nullable=False,
    )

    availability_logistics = Column(
        "fld_availability_logistics",
        Float,
        default=__defaults__["availability_logistics"],
    )
    availability_mission = Column(
        "fld_availability_mission",
        Float,
        default=__defaults__["availability_mission"],
    )
    cost = Column("fld_cost", Float, default=__defaults__["cost"])
    function_code = Column(
        "fld_function_code",
        String(16),
        default=__defaults__["function_code"],
    )
    hazard_rate_logistics = Column(
        "fld_hazard_rate_logistics",
        Float,
        default=__defaults__["hazard_rate_logistics"],
    )
    hazard_rate_mission = Column(
        "fld_hazard_rate_mission", Float, default=__defaults__["hazard_rate_mission"]
    )
    level = Column("fld_level", Integer, default=__defaults__["level"])
    mmt = Column("fld_mmt", Float, default=__defaults__["mmt"])
    mcmt = Column("fld_mcmt", Float, default=__defaults__["mcmt"])
    mpmt = Column("fld_mpmt", Float, default=__defaults__["mpmt"])
    mtbf_logistics = Column(
        "fld_mtbf_logistics", Float, default=__defaults__["mtbf_logistics"]
    )
    mtbf_mission = Column(
        "fld_mtbf_mission", Float, default=__defaults__["mtbf_mission"]
    )
    mttr = Column("fld_mttr", Float, default=__defaults__["mttr"])
    name = Column("fld_name", String(256), default=__defaults__["name"])
    parent_id = Column("fld_parent_id", Integer, default=__defaults__["parent_id"])
    remarks = Column("fld_remarks", String, default=__defaults__["remarks"])
    safety_critical = Column(
        "fld_safety_critical", Integer, default=__defaults__["safety_critical"]
    )
    total_mode_count = Column(
        "fld_total_mode_count", Integer, default=__defaults__["total_mode_count"]
    )
    total_part_count = Column(
        "fld_total_part_count", Integer, default=__defaults__["total_part_count"]
    )
    type_id = Column("fld_type_id", Integer, default=__defaults__["type_id"])

    # Define the relationships to other tables in the RAMSTK Program database.
    hazard: relationship = relationship(
        "RAMSTKHazardRecord",
        backref="function",
        passive_deletes=True,
    )

    def get_attributes(self):
        """Retrieve current values of the RAMSTKFunction data model attributes.

        :return: {revision_id, function_id, availability_logistics,
                  availability_mission, cost, function_code,
                  hazard_rate_logistics, hazard_rate_mission, level, mmt, mcmt,
                  mpmt, mtbf_logistics, mtbf_mission, mttr, name, parent_id,
                  remarks, safety_critical, total_mode_count, total_part_count,
                  type_id} pairs.
        :rtype: tuple
        """
        _values = {
            "revision_id": self.revision_id,
            "function_id": self.function_id,
            "availability_logistics": self.availability_logistics,
            "availability_mission": self.availability_mission,
            "cost": self.cost,
            "function_code": self.function_code,
            "hazard_rate_logistics": self.hazard_rate_logistics,
            "hazard_rate_mission": self.hazard_rate_mission,
            "level": self.level,
            "mmt": self.mmt,
            "mcmt": self.mcmt,
            "mpmt": self.mpmt,
            "mtbf_logistics": self.mtbf_logistics,
            "mtbf_mission": self.mtbf_mission,
            "mttr": self.mttr,
            "name": self.name,
            "parent_id": self.parent_id,
            "remarks": self.remarks,
            "safety_critical": self.safety_critical,
            "total_mode_count": self.total_mode_count,
            "total_part_count": self.total_part_count,
            "type_id": self.type_id,
        }

        return _values



# -*- coding: utf-8 -*-
#
#       ramstk.models.commondb.RAMSTKUser.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2019 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTKUser Table Module."""

# Third Party Imports
from sqlalchemy import Column, Integer, String

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord


class RAMSTKUser(RAMSTK_BASE, RAMSTKBaseRecord):
    """Class to represent the table ramstk_user in the RAMSTK Common database.

    This table shares a One-to-Many relationship with ramstk_workgroup.
    """

    __defaults__ = {
        "user_lname": "Last Name",
        "user_fname": "First Name",
        "user_email": "EMail",
        "user_phone": "867.5309",
        "user_group_id": "0",
    }
    __tablename__ = "ramstk_user"
    __table_args__ = {"extend_existing": True}

    user_id = Column(
        "fld_user_id",
        Integer,
        primary_key=True,
        autoincrement=True,
        nullable=False,
    )
    user_lname = Column(
        "fld_user_lname", String(256), default=__defaults__["user_lname"]
    )
    user_fname = Column(
        "fld_user_fname", String(256), default=__defaults__["user_fname"]
    )
    user_email = Column(
        "fld_user_email", String(256), default=__defaults__["user_email"]
    )
    user_phone = Column(
        "fld_user_phone", String(256), default=__defaults__["user_phone"]
    )
    user_group_id = Column(
        "fld_user_group_id", String(256), default=__defaults__["user_group_id"]
    )

    def get_attributes(self):
        """Retrieve the current values of the RAMSTKUser data model attributes.

        :return: {user_id, user_lname, user_fname, user_email, user_phone,
                  user_group_id} pairs.
        :rtype: dict
        """
        _attributes = {
            "user_id": self.user_id,
            "user_lname": self.user_lname,
            "user_fname": self.user_fname,
            "user_email": self.user_email,
            "user_phone": self.user_phone,
            "user_group_id": self.user_group_id,
        }

        return _attributes



# type: ignore
# -*- coding: utf-8 -*-
#
#       ramstk.analyses.prediction.Connection.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2017 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Connection MIL-HDBK-217F Constants and Calculations Module."""

# Standard Library Imports
from math import exp
from typing import Any, Dict

PART_COUNT_LAMBDA_B = {
    1: {
        1: [
            0.011,
            0.14,
            0.11,
            0.069,
            0.20,
            0.058,
            0.098,
            0.23,
            0.34,
            0.37,
            0.0054,
            0.16,
            0.42,
            6.8,
        ],
        2: [
            0.012,
            0.015,
            0.13,
            0.075,
            0.21,
            0.06,
            0.1,
            0.22,
            0.32,
            0.38,
            0.0061,
            0.18,
            0.54,
            7.3,
        ],
    },
    2: [
        0.0054,
        0.021,
        0.055,
        0.035,
        0.10,
        0.059,
        0.11,
        0.085,
        0.16,
        0.19,
        0.0027,
        0.078,
        0.21,
        3.4,
    ],
    3: [
        0.0019,
        0.0058,
        0.027,
        0.012,
        0.035,
        0.015,
        0.023,
        0.021,
        0.025,
        0.048,
        0.00097,
        0.027,
        0.070,
        1.3,
    ],
    4: [
        0.053,
        0.11,
        0.37,
        0.69,
        0.27,
        0.27,
        0.43,
        0.85,
        1.5,
        1.0,
        0.027,
        0.53,
        1.4,
        27.0,
    ],
    5: {
        1: [
            0.0026,
            0.0052,
            0.018,
            0.010,
            0.029,
            0.010,
            0.016,
            0.016,
            0.021,
            0.042,
            0.0013,
            0.023,
            0.062,
            1.1,
        ],
        2: [
            0.00014,
            0.00028,
            0.00096,
            0.00056,
            0.0015,
            0.00056,
            0.00084,
            0.00084,
            0.0011,
            0.0022,
            0.00007,
            0.0013,
            0.0034,
            0.059,
        ],
        3: [
            0.00026,
            0.00052,
            0.0018,
            0.0010,
            0.0029,
            0.0010,
            0.0016,
            0.0016,
            0.0021,
            0.0042,
            0.00013,
            0.0023,
            0.0062,
            0.11,
        ],
        4: [
            0.000050,
            0.000100,
            0.000350,
            0.000200,
            0.000550,
            0.000200,
            0.000300,
            0.000300,
            0.000400,
            0.000800,
            0.000025,
            0.000450,
            0.001200,
            0.021000,
        ],
        5: [
            0.0000035,
            0.000007,
            0.000025,
            0.000014,
            0.000039,
            0.000014,
            0.000021,
            0.000021,
            0.000028,
            0.000056,
            0.0000018,
            0.000031,
            0.000084,
            0.0015,
        ],
        6: [
            0.00012,
            0.00024,
            0.00084,
            0.00048,
            0.0013,
            0.00048,
            0.00072,
            0.00072,
            0.00096,
            0.0019,
            0.00005,
            0.0011,
            0.0029,
            0.050,
        ],
        7: [
            0.000069,
            0.000138,
            0.000483,
            0.000276,
            0.000759,
            0.000276,
            0.000414,
            0.000414,
            0.000552,
            0.001104,
            0.000035,
            0.000621,
            0.001656,
            0.02898,
        ],
    },
}
PART_COUNT_PI_Q = [1.0, 2.0]
PART_STRESS_LAMBDA_B = {
    4: [0.000041, 0.00026],
    5: [0.0026, 0.00014, 0.00026, 0.00005, 0.0000035, 0.00012, 0.000069],
}
PART_STRESS_PI_Q = {4: [1.0, 2.0], 5: [1.0, 1.0, 2.0, 20.0]}
PI_E = {
    1: {
        1: [
            1.0,
            1.0,
            8.0,
            5.0,
            13.0,
            3.0,
            5.0,
            8.0,
            12.0,
            19.0,
            0.5,
            10.0,
            27.0,
            490.0,
        ],
        2: [
            2.0,
            5.0,
            21.0,
            10.0,
            27.0,
            12.0,
            18.0,
            17.0,
            25.0,
            37.0,
            0.8,
            20.0,
            54.0,
            970.0,
        ],
    },
    2: {
        1: [
            1.0,
            3.0,
            8.0,
            5.0,
            13.0,
            6.0,
            11.0,
            6.0,
            11.0,
            19.0,
            0.5,
            10.0,
            27.0,
            490.0,
        ],
        2: [
            2.0,
            7.0,
            17.0,
            10.0,
            26.0,
            14.0,
            22.0,
            14.0,
            22.0,
            37.0,
            0.8,
            20.0,
            54.0,
            970.0,
        ],
    },
    3: [1.0, 3.0, 14.0, 6.0, 18.0, 8.0, 12.0, 11.0, 13.0, 25.0, 0.5, 14.0, 36.0, 650.0],
    4: [1.0, 2.0, 7.0, 5.0, 13.0, 5.0, 8.0, 16.0, 28.0, 19.0, 0.5, 10.0, 27.0, 500.0],
    5: [1.0, 2.0, 7.0, 4.0, 11.0, 4.0, 6.0, 6.0, 8.0, 16.0, 0.5, 9.0, 24.0, 420.0],
}
PI_K = [1.0, 1.5, 2.0, 3.0, 4.0]
REF_TEMPS = {1: 473.0, 2: 423.0, 3: 373.0, 4: 358.0, 5: 423.0}


def calculate_active_pins_factor(n_active_pins: int) -> float:
    """Calculate the active pins factor (piP).

    :param n_active_pins: the number of active pins in the connector.
    :return: _pi_p; the calculated value of piP.
    :rtype: float
    """
    return exp(((n_active_pins - 1) / 10.0) ** 0.51064)


def calculate_complexity_factor(n_circuit_planes: int) -> float:
    """Calculate the complexity factor (piC).

    :param n_circuit_planes: the number of planes in the PCB/PWA.
    :return: _pi_c; the calculated value of the complexity factor.
    :rtype: float
    """
    if n_circuit_planes > 2:
        _pi_c = 0.65 * n_circuit_planes ** 0.63
    else:
        _pi_c = 1.0

    return _pi_c


def calculate_insert_temperature(contact_gauge: int, current_operating: float) -> float:
    """Calculate the insert temperature.

    Operating current can be passed as float or integer:
    >>> calculate_insert_temperature(1, 16, 0.05)
    0.0010736063482992093
    >>> calculate_insert_temperature(1, 16, 5)
    5.380777957087587

    A KeyError is raised if the contact gauge are unknown:
    >>> calculate_insert_temperature(1, 6, 0.05)
    Traceback (most recent call last):
        ...
    KeyError: 6

    A TypeError is raised if the operating current is passed as a string:
    >>> calculate_insert_temperature(1, 16, '0.05')
    Traceback (most recent call last):
        ...
    TypeError: unsupported operand type(s) for ** or pow(): 'str' and 'float'

    :param contact_gauge: the standard gauge of the connection contact.
    :param current_operating: the nominal current carried by each
        connection contact.
    :return: _temperature_rise; the calculated temperature of the connection's
        insert.
    :rtype: float
    :raise: KeyError when an unknown contact gauge is passed.
    :raise: TypeError when the operating current is passed as a string.
    """
    _dic_factors = {12: 0.1, 16: 0.274, 20: 0.64, 22: 0.989, 26: 2.1}

    _fo = _dic_factors[contact_gauge]
    _temperature_rise = _fo * current_operating ** 1.85

    return _temperature_rise


def calculate_part_count(**attributes: Dict[str, Any]) -> float:
    """Wrap get_part_count_lambda_b().

    This wrapper allows us to pass an attributes dict from a generic parts
    count function.

    :param attributes: the attributes for the connection being calculated.
    :return: _base_hr; the parts count base hazard rates.
    :rtype: float
    """
    return get_part_count_lambda_b(
        subcategory_id=attributes["subcategory_id"],
        environment_active_id=attributes["environment_active_id"],
        type_id=attributes["type_id"],
    )


def calculate_part_stress(**attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate the part stress active hazard rate for a connection.

    This function calculates the MIL-HDBK-217F hazard rate using the part
    stress method.

    :param attributes: the attributes for the connection being calculated.
    :return: attributes; the keyword argument (hardware attribute)
        dictionary with updated values.
    :rtype: dict
    """
    attributes["temperature_rise"] = calculate_insert_temperature(
        attributes["contact_gauge"], attributes["current_operating"]
    )
    attributes["piC"] = calculate_complexity_factor(attributes["n_circuit_planes"])
    attributes["piP"] = calculate_active_pins_factor(attributes["n_active_pins"])
    attributes["piK"] = get_mate_unmate_factor(attributes["n_cycles"])

    if attributes["subcategory_id"] == 1:
        _factor_key = get_factor_key(
            attributes["type_id"],
            attributes["specification_id"],
            attributes["insert_id"],
        )
    else:
        _factor_key = 5
    _contact_temp = (
        attributes["temperature_active"] + attributes["temperature_rise"] + 273.0
    )

    attributes["lambda_b"] = calculate_part_stress_lambda_b(
        attributes["subcategory_id"], attributes["type_id"], _contact_temp, _factor_key
    )

    attributes["hazard_rate_active"] = attributes["lambda_b"] * attributes["piE"]
    if attributes["subcategory_id"] == 3:
        attributes["hazard_rate_active"] = (
            attributes["hazard_rate_active"] * attributes["piP"]
        )
    elif attributes["subcategory_id"] == 4:
        attributes["hazard_rate_active"] = (
            attributes["hazard_rate_active"]
            * (
                attributes["n_wave_soldered"] * attributes["piC"]
                + attributes["n_hand_soldered"] * (attributes["piC"] + 13.0)
            )
            * attributes["piQ"]
        )
    elif attributes["subcategory_id"] == 5:
        attributes["hazard_rate_active"] = (
            attributes["hazard_rate_active"] * attributes["piQ"]
        )
    else:
        attributes["hazard_rate_active"] = (
            attributes["hazard_rate_active"] * attributes["piK"] * attributes["piP"]
        )

    return attributes


def calculate_part_stress_lambda_b(
    subcategory_id: int, type_id: int, contact_temperature: float, factor_key: int
) -> float:
    """Calculate part stress base hazard rate (lambda b) from MIL-HDBK-217F.

    This function calculates the MIL-HDBK-217F base hazard rate for the parts
    stress method.

    .. important:: the contact temperature must be calculated by the calling
        function as it is not an attribute of a Connection.

    :param subcategory_id: the subcategory identifier.
    :param type_id: the connection type identifier.
    :param contact_temperature: the operating temperature of the
        contacts.
    :return: _base_hr; the calculates base hazard rate.
    :rtype: float
    :raise: IndexError if passed an unknown type ID.
    :raise: ZeroDivisionError if passed contact temperature = 0.0.
    """
    # Factors are used to calculate base hazard rate for circular/rack and
    # panel connectors.  Key is from dictionary above (1 - 4) or contact
    # gauge (22 - 12).
    _dic_factors = {
        1: [0.2, -1592.0, 5.36],
        2: [0.431, -2073.6, 4.66],
        3: [0.19, -1298.0, 4.25],
        4: [0.77, -1528.8, 4.72],
        5: [0.216, -2073.6, 4.66],
    }

    _ref_temp = REF_TEMPS[factor_key]
    _f0 = _dic_factors[factor_key][0]
    _f1 = _dic_factors[factor_key][1]
    _f2 = _dic_factors[factor_key][2]

    if subcategory_id in [4, 5]:
        _lambda_b = PART_STRESS_LAMBDA_B[subcategory_id][type_id - 1]
    elif subcategory_id == 3:
        _lambda_b = 0.00042
    else:
        _lambda_b = _f0 * exp(
            (_f1 / contact_temperature) + (contact_temperature / _ref_temp) ** _f2
        )

    return _lambda_b


def get_factor_key(type_id: int, specification_id: int, insert_id: int) -> int:
    """Retrieve the reference temperature key for the connection.

    :param subcategory_id: the subcategory identifier.
    :param type_id: the connection type identifier.
    :param specification_id: the connection governing specification
        identifier.
    :param insert_id: the insert material identifier.
    :return: _key; the key to use to select the reference temperature and other
        factors.
    :rtype: int
    """
    # Reference temperature is used to calculate base hazard rate for
    # circular/rack and panel connectors.  To get the reference temperature
    # dictionary key, we quesry the key dictionary in which the first key is
    # the connector type ID, second key is the specification ID.  The insert
    # material ID is the index in the list returned.
    _dic_keys = {
        1: {
            1: [2, 2, 2, 2, 2, 2],
            2: [2, 2, 2, 2, 2, 2],
            3: [1, 1, 1, 2, 2, 2, 2, 2, 2],
            4: [1, 1, 1, 2, 2, 2, 2, 2, 2],
            5: [1, 1, 1, 2, 2, 2, 2, 2, 2],
        },
        2: {
            1: [2, 2, 2, 2, 2, 2, 4, 4, 4],
            2: [1, 1, 1, 2, 2, 2, 2, 2, 2, 4, 4, 4],
            3: [1, 1, 1, 2, 2, 2, 2, 2, 2],
            4: [1, 1, 1, 2, 2, 2, 2, 2, 2],
            5: [2, 2, 2, 2, 2, 2],
            6: [2, 2, 2, 2, 2, 2],
        },
        3: {1: [2, 2, 2, 2, 2, 2, 4, 4, 4], 2: [2, 2, 2, 2, 2, 2, 4, 4, 4]},
        4: {
            1: [3, 3],
            2: [3, 3],
            3: [3, 3],
            4: [3, 3],
            5: [3, 3],
            6: [3, 3],
            7: [3, 3],
            8: [3, 3, 2, 2, 2, 2, 2, 2],
        },
        5: {1: [3, 3, 2, 2, 2, 2, 2, 2]},
    }
    return _dic_keys[type_id][specification_id][insert_id - 1]


def get_mate_unmate_factor(n_cycles: float) -> float:
    """Retrieve the mating/unmating factor (piK).

    :param n_cycles: the average number of mate/unmate cycles expected
        per hour of operation.
    :return: _pi_k; the mate_unmate_factor.
    :rtype: float
    """
    if n_cycles <= 0.05:
        _pi_k = PI_K[0]
    elif 0.05 < n_cycles <= 0.5:
        _pi_k = PI_K[1]
    elif 0.5 < n_cycles <= 5.0:
        _pi_k = PI_K[2]
    elif 5.0 < n_cycles <= 50.0:
        _pi_k = PI_K[3]
    else:
        _pi_k = PI_K[4]

    return _pi_k


def get_part_count_lambda_b(**kwargs: Dict[str, int]) -> float:
    """Retrieve the parts count base hazard rate (lambda b) from MIL-HDBK-217F.

    This function retrieves the MIL-HDBK-217F parts count base hazard rate.
    The dictionary PART_COUNT_LAMBDA_B contains the MIL-HDBK-217F parts count
    base hazard rates.  Keys are for PART_COUNT_LAMBDA_B are:

        #. subcategory_id
        #. environment_active_id
        #. type id; if the connection subcategory is NOT type dependent, then
            the second key will be zero.

    Current subcategory IDs are:

    +----------------+-------------------------------+-----------------+
    | Subcategory    |           Connection          | MIL-HDBK-217F   |
    |       ID       |              Style            |    Section      |
    +================+===============================+=================+
    |        1       | Circular, Rack and Panel,     |       15.1      |
    |                | Coaxial, Triaxial             |                 |
    +----------------+-------------------------------+-----------------+
    |        2       | PCB/PWA Edge                  |       15.2      |
    +----------------+-------------------------------+-----------------+
    |        3       | IC Socket                     |       15.3      |
    +----------------+-------------------------------+-----------------+
    |        4       | Plated Through Hole (PTH)     |       16.1      |
    +----------------+-------------------------------+-----------------+
    |        5       | Non-PTH                       |       17.1      |
    +----------------+-------------------------------+-----------------+

    :param id_keys: the ID's used as keys when selecting
        the base hazard rate.  The keys are subcategory_id,
        environment_active_id, and type_id.
    :return: _base_hr; the parts count base hazard rate.
    :rtype: float
    :raise: KeyError if passed an unknown subcategory ID or type ID.
    :raise: IndexError if passed an unknown active environment ID.
    """
    _subcategory_id = kwargs.get("subcategory_id", 0)
    _type_id = kwargs.get("type_id", 0)
    _environment_active_id = kwargs.get("environment_active_id", 0)

    if _subcategory_id in [1, 5]:
        _base_hr = PART_COUNT_LAMBDA_B[_subcategory_id][_type_id][
            _environment_active_id - 1
        ]
    else:
        _base_hr = PART_COUNT_LAMBDA_B[_subcategory_id][_environment_active_id - 1]

    return _base_hr



# type: ignore
# -*- coding: utf-8 -*-
#
#       ramstk.analyses.statistics.lognormal.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Exponential Module."""

# Standard Library Imports
from typing import Tuple

# Third Party Imports
import scipy
from scipy.stats import lognorm


def get_hazard_rate(
    shape: float, time: float, location: float = 0.0, scale: float = 1.0
) -> float:
    """Calculates the hazard rate given a scale, shape, and location parameter.

    This function calculates the rate parameter given the scale, shape, time, and,
    optionally, a location parameter.

        >>> get_hazard_rate(0.9663, 4, scale=33.65)
        0.6610467490673426

        >>> get_hazard_rate(0.9663, 4, location=1.85, scale=33.65)
        1.5117773383839221

        >>> get_hazard_rate(0, 4, scale=33.65)
        nan

        >>> get_hazard_rate(0.9663, 0, scale=33.65)
        0.0

    The relationship to scipy parameters is:
        sigma = s
        mu = log(scale)

    :param scale: the value of the scale parameter.
    :param shape: the value of the shape parameter.
    :param time: the time at which to calculate the hazard rate.
    :param location: the value of the location parameter.
    :return: _hazard_rate; the hazard rate.
    :rtype: float
    """
    if time <= 0.0:
        _hazard_rate = 0.0
    else:
        _hazard_rate = lognorm.pdf(
            time, shape, loc=location, scale=scale
        ) / lognorm.cdf(time, shape, loc=location, scale=scale)

    return _hazard_rate


def get_mtbf(shape: float, location: float = 0.0, scale: float = 1.0) -> float:
    """Calculate the MTBF given a shape (sigma) and scale (mu) parameter.

        >>> get_mtbf(0.9663, scale=33.65)
        53.671433807942236

        >>> get_mtbf(0.0, scale=33.65)
        nan

    :param shape: the shape parameter.
    :param location: the location parameter.
    :param scale: the scale parameter.
    :return: _mtbf; the MTBF.
    :rtype: float
    """
    return lognorm.mean(
        shape,
        loc=location,
        scale=scale,
    )


def get_survival(
    shape: float,
    time: float,
    location: float = 0.0,
    scale: float = 1.0,
) -> float:
    """Calculate value of the survival function at time given scale and location.

    This function returns the lower and upper alpha bounds as well as the point
    estimate of the survival function at time.

        >>> get_survival(0.9663, 5.0, scale=33.65)
        0.9757561

        >>> get_survival(0.9663, 5.0, location=1.85, scale=33.65)
        0.9928813

    :param scale: the point estimate of the scale parameter.
    :param time: the time at which to calculate the survival function.
    :param location: the point estimate of the location parameter.
    :return: _surv; the value of the survival function at time.
    :rtype: float
    """
    return lognorm.sf(time, shape, loc=location, scale=scale)


def do_fit(data, **kwargs) -> Tuple[float, float, float]:
    """Fits the provided data to the LOGN distribution and estimates scale and
    location.

    :param data: the data to use in the fit.
    :return: (_shape, _location, _scale); the estimated parameters.
    :rtype: tuple
    """
    _location = kwargs.get("location", 0.0)  # Initial guess for location.
    _floc = kwargs.get("floc", None)  # Value to fix location parameter.
    _scale = kwargs.get("scale", 0.0)  # Initial guess for scale.
    _method = kwargs.get("method", "MLE")  # One of MLE or MM.

    if _floc is None:
        if scipy.__version__ >= "1.7.1":
            _shape, _location, _scale = lognorm.fit(
                data,
                loc=_location,
                scale=_scale,
                method=_method,
            )
        else:
            _shape, _location, _scale = lognorm.fit(
                data,
                loc=_location,
                scale=_scale,
            )
    else:
        if scipy.__version__ >= "1.7.1":
            _shape, _location, _scale = lognorm.fit(
                data,
                loc=_location,
                scale=_scale,
                floc=_floc,
                method=_method,
            )
        else:
            _shape, _location, _scale = lognorm.fit(
                data,
                loc=_location,
                scale=_scale,
                floc=_floc,
            )

    return _shape, _location, _scale



# pylint: disable=unused-import
# -*- coding: utf-8 -*-
#
#       ramstk.views.__init__.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2007-2020 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTK views package."""



# Standard Library Imports
from typing import Any

# RAMSTK Package Imports
from ramstk.models import RAMSTKBaseTable as RAMSTKBaseTable
from ramstk.models import RAMSTKSiteInfoRecord as RAMSTKSiteInfoRecord

class RAMSTKSiteInfoTable(RAMSTKBaseTable):
    _db_id_colname: str
    _db_tablename: str
    _tag: str
    _lst_id_columns: Any
    _record: Any
    pkey: str
    def __init__(self, **kwargs) -> None: ...
